# Copyright (c) 2026, NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
#
#
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

# distutils: language = c++
# cython: language_level=3
# cython: linetrace=True

# Python signatures are declared in the companion Python stub file _bindings.pyi
# Make sure to update PYI with change to Python API to ensure that Python
# static type checker tools like mypy green-lights cuda.compute

from libc.string cimport memset, memcpy
from libc.stdint cimport uint8_t, uint32_t, uint64_t, int64_t, uintptr_t
from libc.stdlib cimport malloc, free
from cpython.bytes cimport PyBytes_FromStringAndSize

from cpython.buffer cimport (
    Py_buffer, PyBUF_SIMPLE, PyBUF_ANY_CONTIGUOUS,
    PyBuffer_Release, PyObject_CheckBuffer, PyObject_GetBuffer
)
from cpython.pycapsule cimport (
    PyCapsule_CheckExact, PyCapsule_IsValid, PyCapsule_GetPointer
)

import ctypes
from enum import IntEnum
cdef extern from "<cuda.h>":
    cdef struct OpaqueCUstream_st
    cdef struct OpaqueCUkernel_st
    cdef struct OpaqueCUlibrary_st

    ctypedef int CUresult
    ctypedef OpaqueCUstream_st *CUstream
    ctypedef OpaqueCUkernel_st *CUkernel
    ctypedef OpaqueCUlibrary_st *CUlibrary


cdef extern from "cccl/c/types.h":
    cpdef enum cccl_type_enum:
        INT8 "CCCL_INT8"
        INT16 "CCCL_INT16"
        INT32 "CCCL_INT32"
        INT64 "CCCL_INT64"
        UINT8 "CCCL_UINT8"
        UINT16 "CCCL_UINT16"
        UINT32 "CCCL_UINT32"
        UINT64 "CCCL_UINT64"
        FLOAT16 "CCCL_FLOAT16"
        FLOAT32 "CCCL_FLOAT32"
        FLOAT64 "CCCL_FLOAT64"
        STORAGE "CCCL_STORAGE"
        BOOLEAN "CCCL_BOOLEAN"

    cpdef enum cccl_op_kind_t:
       STATELESS "CCCL_STATELESS"
       STATEFUL "CCCL_STATEFUL"
       PLUS "CCCL_PLUS"
       MINUS "CCCL_MINUS"
       MULTIPLIES "CCCL_MULTIPLIES"
       DIVIDES "CCCL_DIVIDES"
       MODULUS "CCCL_MODULUS"
       EQUAL_TO "CCCL_EQUAL_TO"
       NOT_EQUAL_TO "CCCL_NOT_EQUAL_TO"
       GREATER "CCCL_GREATER"
       LESS "CCCL_LESS"
       GREATER_EQUAL "CCCL_GREATER_EQUAL"
       LESS_EQUAL "CCCL_LESS_EQUAL"
       LOGICAL_AND "CCCL_LOGICAL_AND"
       LOGICAL_OR "CCCL_LOGICAL_OR"
       LOGICAL_NOT "CCCL_LOGICAL_NOT"
       BIT_AND "CCCL_BIT_AND"
       BIT_OR "CCCL_BIT_OR"
       BIT_XOR "CCCL_BIT_XOR"
       BIT_NOT "CCCL_BIT_NOT"
       IDENTITY "CCCL_IDENTITY"
       NEGATE "CCCL_NEGATE"
       MINIMUM "CCCL_MINIMUM"
       MAXIMUM "CCCL_MAXIMUM"

    cpdef enum cccl_iterator_kind_t:
       POINTER "CCCL_POINTER"
       ITERATOR "CCCL_ITERATOR"

    cdef struct cccl_type_info:
        size_t size
        size_t alignment
        cccl_type_enum type

    cdef enum cccl_op_code_type:
        CCCL_OP_LTOIR
        CCCL_OP_CPP_SOURCE

    cdef struct cccl_op_t:
        cccl_op_kind_t type
        const char* name
        const char* code
        size_t code_size
        cccl_op_code_type code_type
        size_t size
        size_t alignment
        void *state
        const char** extra_ltoirs
        size_t* extra_ltoir_sizes
        size_t num_extra_ltoirs

    cdef struct cccl_value_t:
        cccl_type_info type
        void *state

    cdef union cccl_increment_t:
        int64_t signed_offset
        uint64_t unsigned_offset

    ctypedef void (*cccl_host_op_fn_ptr_t)(void *, cccl_increment_t) nogil

    cdef struct cccl_iterator_t:
        size_t size
        size_t alignment
        cccl_iterator_kind_t type
        cccl_op_t advance
        cccl_op_t dereference
        cccl_type_info value_type
        void *state
        cccl_host_op_fn_ptr_t host_advance

    cpdef enum cccl_sort_order_t:
        ASCENDING "CCCL_ASCENDING"
        DESCENDING "CCCL_DESCENDING"

    cpdef enum cccl_init_kind_t:
        VALUE_INIT "CCCL_VALUE_INIT"
        FUTURE_VALUE_INIT "CCCL_FUTURE_VALUE_INIT"
        NO_INIT "CCCL_NO_INIT"

    cpdef enum cccl_determinism_t:
        NOT_GUARANTEED "CCCL_NOT_GUARANTEED"
        RUN_TO_RUN "CCCL_RUN_TO_RUN"
        GPU_TO_GPU "CCCL_GPU_TO_GPU"

    cpdef enum cccl_binary_search_mode_t:
        LOWER_BOUND "CCCL_BINARY_SEARCH_LOWER_BOUND"
        UPPER_BOUND "CCCL_BINARY_SEARCH_UPPER_BOUND"

cdef void arg_type_check(
    str arg_name,
    object expected_type,
    object arg
) except *:
    if not isinstance(arg, expected_type):
        raise TypeError(
            f"Expected {arg_name} to have type '{expected_type}', "
            f"got '{type(arg)}'"
        )

OpKind = cccl_op_kind_t
TypeEnum = cccl_type_enum
IteratorKind = cccl_iterator_kind_t
SortOrder = cccl_sort_order_t
InitKind = cccl_init_kind_t
Determinism = cccl_determinism_t
BinarySearchMode = cccl_binary_search_mode_t

cdef void _validate_alignment(int alignment) except *:
    """
    Alignment must be positive integer and a power of two
    that can be represented by uint32_t type.
    """
    cdef uint32_t val
    if alignment < 1:
        raise ValueError(
            "Alignment must be non-negative, "
            f"got {alignment}."
        )
    val = <uint32_t>alignment
    if (val & (val - 1)) != 0:
        raise ValueError(
            "Alignment must be a power of two, "
            f"got {alignment}"
        )


cdef class Op:
    """
    Represents CCCL Operation

    Args:
        name (str):
            Name of the operation
        operator_type (OpKind):
            Whether operator is stateless or stateful
        ltoir (bytes):
            The LTOIR for the operation compiled for device
        state (bytes, optional):
            State for the stateful operation.
        state_alignment (int, optional):
            Alignment of the state struct. Default: `1`.
        extra_ltoirs (list of bytes, optional):
            Additional LTOIR modules to link with this operation.
    """
    # need Python owner of memory used for operator name
    cdef bytes op_encoded_name
    cdef bytes code_bytes
    cdef bytes state_bytes
    cdef list extra_ltoirs_list  # Python list to keep bytes alive
    cdef const char** extra_ltoirs_ptrs
    cdef size_t* extra_ltoir_sizes_arr
    cdef cccl_op_t op_data


    cdef void _set_members(self, cccl_op_kind_t op_type, str name, bytes lto_ir, bytes state, int state_alignment, list extra_ltoirs):
        memset(&self.op_data, 0, sizeof(cccl_op_t))
        # Reference Python objects in the class to ensure lifetime
        self.op_encoded_name = name.encode("utf-8")
        self.code_bytes = lto_ir
        self.state_bytes = state
        self.extra_ltoirs_list = extra_ltoirs if extra_ltoirs else []
        # set fields of op_data struct
        self.op_data.type = op_type
        self.op_data.name = <const char *>self.op_encoded_name
        self.op_data.code = <const char *>lto_ir
        self.op_data.code_size = len(lto_ir)
        self.op_data.code_type = cccl_op_code_type.CCCL_OP_LTOIR
        self.op_data.size = len(state)
        self.op_data.alignment = state_alignment
        self.op_data.state = <void *><const char *>state

        # Handle extra_ltoirs
        cdef size_t num_extra = len(self.extra_ltoirs_list)
        if num_extra > 0:
            self.extra_ltoirs_ptrs = <const char**>malloc(num_extra * sizeof(const char*))
            self.extra_ltoir_sizes_arr = <size_t*>malloc(num_extra * sizeof(size_t))
            for i in range(num_extra):
                ltoir_bytes = <bytes>self.extra_ltoirs_list[i]
                self.extra_ltoirs_ptrs[i] = <const char*>ltoir_bytes
                self.extra_ltoir_sizes_arr[i] = len(ltoir_bytes)
            self.op_data.extra_ltoirs = self.extra_ltoirs_ptrs
            self.op_data.extra_ltoir_sizes = self.extra_ltoir_sizes_arr
            self.op_data.num_extra_ltoirs = num_extra
        else:
            self.extra_ltoirs_ptrs = NULL
            self.extra_ltoir_sizes_arr = NULL
            self.op_data.extra_ltoirs = NULL
            self.op_data.extra_ltoir_sizes = NULL
            self.op_data.num_extra_ltoirs = 0


    def __cinit__(self, /, *, name = None, operator_type = None, ltoir = None, state = None, state_alignment = 1, extra_ltoirs = None):
        if name is None and ltoir is None:
            name = ""
            ltoir = b""
        if state is None:
            state = b""
        if operator_type is None:
            operator_type = OpKind.STATELESS
        if extra_ltoirs is None:
            extra_ltoirs = []
        arg_type_check(arg_name="name", expected_type=str, arg=name)
        arg_type_check(arg_name="ltoir", expected_type=bytes, arg=ltoir)
        arg_type_check(arg_name="state", expected_type=bytes, arg=state)
        arg_type_check(arg_name="state_alignment", expected_type=int, arg=state_alignment)
        arg_type_check(arg_name="extra_ltoirs", expected_type=list, arg=extra_ltoirs)
        for i, el in enumerate(extra_ltoirs):
            if not isinstance(el, bytes):
                raise TypeError(f"extra_ltoirs[{i}] must be bytes, got {type(el)}")
        if not isinstance(operator_type, OpKind):
            raise TypeError(
                f"The operator_type argument should be an enumerator of operator kinds"
            )
        _validate_alignment(state_alignment)
        self._set_members(
            <cccl_op_kind_t> operator_type.value,
            <str> name,
            <bytes> ltoir,
            <bytes> state,
            <int> state_alignment,
            <list> extra_ltoirs
        )

    def __dealloc__(self):
        if self.extra_ltoirs_ptrs != NULL:
            free(self.extra_ltoirs_ptrs)
            self.extra_ltoirs_ptrs = NULL
        if self.extra_ltoir_sizes_arr != NULL:
            free(self.extra_ltoir_sizes_arr)
            self.extra_ltoir_sizes_arr = NULL

    cdef void set_state(self, bytes state):
        self.state_bytes = state
        self.op_data.state = <void *><const char *>state

    @property
    def state(self):
       return self.state_bytes

    @state.setter
    def state(self, bytes new_value):
        self.set_state(<bytes>new_value)

    @property
    def name(self):
        return self.op_encoded_name.decode("utf-8")

    @property
    def ltoir(self):
        # Backward compatibility property
        return self.code_bytes

    @property
    def code(self):
        return self.code_bytes

    @property
    def state_alignment(self):
        return self.op_data.alignment

    @property
    def state_typenum(self):
        return self.op_data.type

    @property
    def extra_ltoirs(self):
        return self.extra_ltoirs_list

    def as_bytes(self):
        "Debugging utility to view memory content of library struct"
        cdef uint8_t[:] mem_view = bytearray(sizeof(self.op_data))
        memcpy(&mem_view[0], &self.op_data, sizeof(self.op_data))
        return bytes(mem_view)


cdef class TypeInfo:
    """
    Represents CCCL type info structure

    Args:
        size (int):
            Size of the type in bytes.
        alignment (int):
            Alignment of the type in bytes.
        type_enum (TypeEnum):
            Enumeration member identifying the type.
    """
    cdef cccl_type_info type_info

    def __cinit__(self, int size, int alignment, cccl_type_enum type_enum):
        if size < 1:
            raise ValueError("Size argument must be positive")
        _validate_alignment(alignment)
        self.type_info.size = size
        self.type_info.alignment = alignment
        self.type_info.type = type_enum

    @property
    def size(self):
        return self.type_info.size

    @property
    def alignment(self):
        return self.type_info.alignment

    @property
    def typenum(self):
        return self.type_info.type

    def as_bytes(self):
        "Debugging utility to view memory content of library struct"
        cdef uint8_t[:] mem_view = bytearray(sizeof(self.type_info))
        memcpy(&mem_view[0], &self.type_info, sizeof(self.type_info))
        return bytes(mem_view)


cdef class Value:
    """
    Represents CCCL value structure

    Args:
        value_type (TypeInfo):
            type descriptor
        state (object):
            state of the value type. Object is expected to
            implement Python buffer protocol and be able to provide
            simple contiguous array of type `uint8_t`.
    """
    cdef uint8_t[::1] state_obj
    cdef TypeInfo value_type
    cdef cccl_value_t value_data;

    def __cinit__(self, TypeInfo value_type, uint8_t[::1] state):
        self.state_obj = state
        self.value_type = value_type
        self.value_data.type = value_type.type_info
        self.value_data.state = <void *>&state[0]

    @property
    def type(self):
        return self.value_type

    @property
    def state(self):
        return self.state_obj

    @state.setter
    def state(self, uint8_t[::1] new_value):
        if (len(self.state_obj) == len(new_value)):
            self.state_obj = new_value
            self.value_data.state = <void *>&self.state_obj[0]
        else:
            raise ValueError("Size mismatch")

    def as_bytes(self):
        "Debugging utility to view memory of native struct"
        cdef uint8_t[:] mem_view = bytearray(sizeof(self.value_data))
        memcpy(&mem_view[0], &self.value_data, sizeof(self.value_data))
        return bytes(mem_view)


cdef void ensure_buffer(object o) except *:
    if not PyObject_CheckBuffer(o):
        raise TypeError(
            "Object with buffer protocol expected, "
            f"got {type(o)}"
        )


cdef void * get_buffer_pointer(object o, size_t *size):
    cdef int status = 0
    cdef void *ptr = NULL
    cdef Py_buffer view

    status = PyObject_GetBuffer(o, &view, PyBUF_SIMPLE | PyBUF_ANY_CONTIGUOUS)
    if status != 0:  # pragma: no cover
        size[0] = 0
        raise RuntimeError(
            "Can not access simple contiguous buffer"
        )

    ptr = view.buf
    if size is not NULL:
        size[0] = <size_t>view.len
    PyBuffer_Release(&view)

    return ptr


cdef void * ctypes_typed_pointer_payload_ptr(object ctypes_typed_ptr):
    "Get pointer to the value buffer represented by ctypes.pointer(ctypes_val)"
    cdef size_t size = 0
    cdef size_t *ptr_ref = NULL
    ensure_buffer(ctypes_typed_ptr)
    ptr_ref = <size_t *>get_buffer_pointer(ctypes_typed_ptr, &size)
    return <void *>(ptr_ref[0])


cdef void * ctypes_value_ptr(object ctypes_cdata):
    "Get pointer to the value buffer behind ctypes_val"
    cdef size_t size = 0
    ensure_buffer(ctypes_cdata)
    return get_buffer_pointer(ctypes_cdata, &size)


cdef inline void * int_as_ptr(size_t ptr_val):
    return <void *>(ptr_val)


cdef class StateBase:
    cdef void *ptr
    cdef object ref

    def __cinit__(self):
        self.ptr = NULL
        self.ref = None

    cdef inline void set_state(self, void *ptr, object ref):
        self.ptr = ptr
        self.ref = ref

    @property
    def pointer(self):
        return <size_t>self.ptr

    @property
    def reference(self):
        return self.ref


cdef class Pointer(StateBase):
    "Represents the pointer value"

    def __cinit__(self, arg):
        cdef void *ptr
        cdef object ref

        if isinstance(arg, int):
            ptr = int_as_ptr(arg)
            ref = None
        elif isinstance(arg, ctypes._Pointer):
            ptr = ctypes_typed_pointer_payload_ptr(arg)
            ref = arg
        elif isinstance(arg, ctypes.c_void_p):
            ptr = int_as_ptr(arg.value)
            ref = arg
        else:
            raise TypeError(
                "Expect ctypes pointer, integers, or PointerProxy, "
                f"got type {type(arg)}"
            )
        self.set_state(ptr, ref)


def make_pointer_object(ptr, owner):
    cdef Pointer res = Pointer(0)

    if isinstance(ptr, int):
        res.ptr = int_as_ptr(ptr)
    elif isinstance(ptr, ctypes.c_void_p):
        res.ptr = int_as_ptr(ptr.value)
    else:
        raise TypeError(
            "First argument must be an integer, or ctypes.c_void_p, "
            f"got {type(ptr)}"
        )
    res.ref = owner
    return res


cdef class IteratorState(StateBase):
    "Represents blob referenced by pointer"
    cdef size_t state_nbytes

    def __cinit__(self, arg):
        cdef size_t buffer_size = 0
        cdef void *ptr = NULL
        cdef object ref = None

        super().__init__()
        if isinstance(arg, ctypes._Pointer):
            ptr = ctypes_typed_pointer_payload_ptr(arg)
            ref = arg.contents
            self.state_nbytes = ctypes.sizeof(ref)
        elif PyObject_CheckBuffer(arg):
            ptr = get_buffer_pointer(arg, &buffer_size)
            ref = arg
            self.state_nbytes = buffer_size
        else:
            raise TypeError(
                "Expected a ctypes pointer with content, or object of type bytes or bytearray, "
                f"got type {type(arg)}"
            )
        self.set_state(ptr, ref)

    cdef inline size_t get_size(self):
        return self.state_nbytes

    @property
    def size(self):
        return self.state_nbytes

    def __getbuffer__(self, Py_buffer *buffer, int flags):
        cdef Py_ssize_t cast_size = <Py_ssize_t>self.state_nbytes
        buffer.buf = <char *>self.ptr
        buffer.obj = self
        buffer.len = cast_size
        buffer.readonly = 0
        buffer.itemsize = 1
        buffer.format = "B"  # unsigned char
        buffer.ndim = 1
        buffer.shape = <Py_ssize_t *>&self.state_nbytes
        buffer.strides = &buffer.itemsize
        buffer.suboffsets = NULL
        buffer.internal = NULL

    def __releasebuffer__(self, Py_buffer *buffer):
        pass


cdef const char *function_ptr_capsule_name = "void (void *, cccl_increment_t)";

cdef bint is_function_pointer_capsule(object o) noexcept:
    """
    Returns non-zero if input is a valid capsule with
    name 'void (void *, cccl_increment_t)'.
    """
    return (
        PyCapsule_CheckExact(o) and
        PyCapsule_IsValid(o, function_ptr_capsule_name)
    )


cdef inline void* get_function_pointer_from_capsule(object cap) except *:
    return PyCapsule_GetPointer(cap, function_ptr_capsule_name)


cdef cccl_host_op_fn_ptr_t unbox_host_advance_fn(object host_fn_obj) except *:
    cdef void *fn_ptr = NULL
    if isinstance(host_fn_obj, ctypes._CFuncPtr):
        # the _CFuncPtr object encapsulates a pointer to the function pointer
        fn_ptr = ctypes_typed_pointer_payload_ptr(host_fn_obj)
        return <cccl_host_op_fn_ptr_t>fn_ptr

    if isinstance(host_fn_obj, int):
        fn_ptr = <void *><uintptr_t>host_fn_obj
        return <cccl_host_op_fn_ptr_t>fn_ptr

    if isinstance(host_fn_obj, ctypes.c_void_p):
        fn_ptr = <void *><uintptr_t>host_fn_obj.value
        return <cccl_host_op_fn_ptr_t>fn_ptr

    if is_function_pointer_capsule(host_fn_obj):
        fn_ptr = get_function_pointer_from_capsule(host_fn_obj)
        return <cccl_host_op_fn_ptr_t>fn_ptr

    raise TypeError(
        "Expected ctypes function pointer, ctypes.c_void_p, integer or a named capsule, "
        f"got {type(host_fn_obj)}"
    )


cdef class Iterator:
    """
    Represents CCCL iterator.

    Args:
        alignment (int):
            Alignment of the iterator state
        iterator_type (IteratorKind):
            The type of iterator, `IteratorKind.POINTER` or
            `IteratorKind.ITERATOR`
        advance_fn (Op):
            Descriptor for user-defined `advance` function
            compiled for device
        dereference_fn (Op):
            Descriptor for user-defined `dereference` or `assign`
            function compiled for device
        value_type (TypeInfo):
            Descriptor of the type addressed by the iterator
        state (object, optional):
            Python object for the state of the iterator. For iterators of
            type `ITERATOR` the state object is expected to implement Python
            buffer protocol for SIMPLE 1d buffer of type unsigned byte.
            For iterators of type `POINTER` the state may be an integer convertible
            to `uintptr_t`, or a `ctypes` pointer (typed or untyped).
            Value `None` represents absence of iterator state.
        host_advance_fn (object, optional):
            Python object for host callable function to advance state by a given
            increment. The argument may only be set for iterators of type
            `IteratorKind.ITERATOR` and raise an exception otherwise. Supported
            types are `int` or `ctypes.c_void_p` (raw pointer), ctypes function
            pointer, or a Python capsule with name `"void *(void *, cccl_increment_t)"`.
    """
    cdef Op advance
    cdef Op dereference
    cdef object state_obj
    cdef object host_advance_obj
    cdef cccl_iterator_t iter_data

    def __cinit__(self,
        int alignment,
        cccl_iterator_kind_t iterator_type,
        Op advance_fn,
        Op dereference_fn,
        TypeInfo value_type,
        state=None,
        host_advance_fn=None
    ):
        cdef cccl_iterator_kind_t it_kind
        _validate_alignment(alignment)
        it_kind = iterator_type
        if it_kind == cccl_iterator_kind_t.POINTER:
            if state is None:
                self.state_obj = None
                self.iter_data.size = 0
                self.iter_data.state = NULL
            elif isinstance(state, int):
                self.state_obj = None
                self.iter_data.size = 0
                self.iter_data.state = int_as_ptr(state)
            elif isinstance(state, Pointer):
                self.state_obj = state.reference
                self.iter_data.size = 0
                self.iter_data.state = (<Pointer>state).ptr
            else:
                raise TypeError(
                    "Expect for Iterator of kind POINTER, state must have type Pointer or int, "
                    f"got {type(state)}"
                )
            if host_advance_fn is not None:
                raise ValueError(
                    "host_advance_fn must be set to None for iterators of kind POINTER"
                )
            self.iter_data.host_advance = NULL
            self.host_advance_obj = None
        elif it_kind == cccl_iterator_kind_t.ITERATOR:
            if state is None:
                self.state_obj = None
                self.iter_data.size = 0
                self.iter_data.state = NULL
            elif isinstance(state, IteratorState):
                self.state_obj = state.reference
                self.iter_data.size = (<IteratorState>state).size
                self.iter_data.state = (<IteratorState>state).ptr
            else:
                raise TypeError(
                    "For Iterator of kind ITERATOR, state must have type IteratorState, "
                    f"got type {type(state)}"
                )
            if host_advance_fn is not None:
                self.iter_data.host_advance = unbox_host_advance_fn(host_advance_fn)
                self.host_advance_obj = host_advance_fn
            else:
                self.iter_data.host_advance = NULL
                self.host_advance_obj = None
        else:  # pragma: no cover
            raise ValueError("Unrecognized iterator kind")
        self.advance = advance_fn
        self.dereference = dereference_fn
        self.iter_data.alignment = alignment
        self.iter_data.type = <cccl_iterator_kind_t> it_kind
        self.iter_data.advance = self.advance.op_data
        self.iter_data.dereference = self.dereference.op_data
        self.iter_data.value_type = value_type.type_info

    @property
    def advance_op(self):
        return self.advance

    @property
    def dereference_or_assign_op(self):
        return self.dereference

    @property
    def state(self):
        if self.iter_data.type == cccl_iterator_kind_t.POINTER:
            return <size_t>self.iter_data.state
        else:
            return self.state_obj

    @state.setter
    def state(self, new_value):
        cdef ssize_t state_sz = 0
        cdef size_t ptr = 0
        cdef cccl_iterator_kind_t it_kind = self.iter_data.type
        if it_kind == cccl_iterator_kind_t.POINTER:
            if isinstance(new_value, Pointer):
                self.state_obj = (<Pointer>new_value).ref
                self.iter_data.size = state_sz
                self.iter_data.state = (<Pointer>new_value).ptr
            elif isinstance(new_value, int):
                self.state_obj = None
                self.iter_data.size = state_sz
                self.iter_data.state = int_as_ptr(new_value)
            elif new_value is None:
                self.state_obj = None
                self.iter_data.size = 0
                self.iter_data.state = NULL
            else:
                raise TypeError(
                    "For iterator with type POINTER, state value must have type int or type Pointer, "
                    f"got type {type(new_value)}"
                )
        elif it_kind == cccl_iterator_kind_t.ITERATOR:
            if isinstance(new_value, IteratorState):
                self.state_obj = new_value.reference
                self.iter_data.size = (<IteratorState>new_value).size
                self.iter_data.state = (<IteratorState>new_value).ptr
            elif isinstance(new_value, Pointer):
                self.state_obj = new_value.reference
                if self.iter_data.size == 0:
                    raise ValueError("Assigning incomplete state value to iterator without state size information")
                self.iter_data.state = (<Pointer>new_value).ptr
            elif PyObject_CheckBuffer(new_value):
                self.iter_data.state = get_buffer_pointer(new_value, &self.iter_data.size)
                self.state_obj = new_value
            elif new_value is None:
                self.state_obj = None
                self.iter_data.size = 0
                self.iter_data.state = NULL
            else:
                raise TypeError(
                    "For iterator with type ITERATOR, state value must have type IteratorState or type bytes, "
                    f"got type {type(new_value)}"
                )
        else:
            raise TypeError("The new value should be an integer for iterators of POINTER kind, and bytes for ITERATOR kind")

    @property
    def type(self):
        cdef cccl_iterator_kind_t it_kind = self.iter_data.type
        if it_kind == cccl_iterator_kind_t.POINTER:
            return IteratorKind.POINTER
        else:
            return IteratorKind.ITERATOR

    @property
    def value_type(self):
        cdef cccl_type_info type_info = self.iter_data.value_type
        return TypeInfo(type_info.size, type_info.alignment, type_info.type)

    def is_kind_pointer(self):
        cdef cccl_iterator_kind_t it_kind = self.iter_data.type
        return (it_kind == cccl_iterator_kind_t.POINTER)

    def is_kind_iterator(self):
        cdef cccl_iterator_kind_t it_kind = self.iter_data.type
        return (it_kind == cccl_iterator_kind_t.ITERATOR)

    def as_bytes(self):
        "Debugging ulitity to get memory view into library struct"
        cdef uint8_t[:] mem_view = bytearray(sizeof(self.iter_data))
        memcpy(&mem_view[0], &self.iter_data, sizeof(self.iter_data))
        return bytes(mem_view)

    @property
    def host_advance_fn(self):
        return self.host_advance_obj

    @host_advance_fn.setter
    def host_advance_fn(self, func):
        if (self.iter_data.type == cccl_iterator_kind_t.ITERATOR):
            if func is not None:
                self.iter_data.host_advance = unbox_host_advance_fn(func)
                self.host_advance_obj = func
            else:
                self.iter_data.host_advance = NULL
                self.host_advance_obj = None
        else:
            raise ValueError


cdef class CommonData:
    cdef int cc_major
    cdef int cc_minor
    cdef bytes encoded_cub_path
    cdef bytes encoded_thrust_path
    cdef bytes encoded_libcudacxx_path
    cdef bytes encoded_ctk_path

    def __cinit__(self, int cc_major, int cc_minor, str cub_path, str thrust_path, str libcudacxx_path, str ctk_path):
        self.cc_major = cc_major
        self.cc_minor = cc_minor
        self.encoded_cub_path = cub_path.encode("utf-8")
        self.encoded_thrust_path = thrust_path.encode("utf-8")
        self.encoded_libcudacxx_path = libcudacxx_path.encode("utf-8")
        self.encoded_ctk_path = ctk_path.encode("utf-8")

    cdef inline int get_cc_major(self):
        return self.cc_major

    cdef inline int get_cc_minor(self):
        return self.cc_minor

    cdef inline const char * cub_path_get_c_str(self):
        return <const char *>self.encoded_cub_path if self.encoded_cub_path else NULL

    cdef inline const char * thrust_path_get_c_str(self):
        return <const char *>self.encoded_thrust_path if self.encoded_thrust_path else NULL

    cdef inline const char * libcudacxx_path_get_c_str(self):
        return <const char *>self.encoded_libcudacxx_path if self.encoded_libcudacxx_path else NULL

    cdef inline const char * ctk_path_get_c_str(self):
        return <const char *>self.encoded_ctk_path if self.encoded_ctk_path else NULL

    @property
    def compute_capability(self):
        return (self.cc_major, self.cc_minor)

    @property
    def cub_path(self):
        return self.encoded_cub_path.decode("utf-8")

    @property
    def ctk_path(self):
        return self.encoded_ctk_path.decode("utf-8")

    @property
    def thrust_path(self):
        return self.encoded_thrust_path.decode("utf-8")

    @property
    def libcudacxx_path(self):
        return self.encoded_libcudacxx_path.decode("utf-8")

# --------------
#   DeviceReduce
# --------------

cdef extern from "cccl/c/reduce.h":
    cdef struct cccl_device_reduce_build_result_t 'cccl_device_reduce_build_result_t':
        const char* cubin
        size_t cubin_size

    cdef CUresult cccl_device_reduce_build(
        cccl_device_reduce_build_result_t*,
        cccl_iterator_t,
        cccl_iterator_t,
        cccl_op_t,
        cccl_value_t,
        cccl_determinism_t,
        int, int, const char*, const char*, const char*, const char*
    ) nogil

    cdef CUresult cccl_device_reduce(
        cccl_device_reduce_build_result_t,
        void *,
        size_t *,
        cccl_iterator_t,
        cccl_iterator_t,
        uint64_t,
        cccl_op_t,
        cccl_value_t,
        CUstream
    ) nogil

    cdef CUresult cccl_device_reduce_nondeterministic(
        cccl_device_reduce_build_result_t,
        void *,
        size_t *,
        cccl_iterator_t,
        cccl_iterator_t,
        uint64_t,
        cccl_op_t,
        cccl_value_t,
        CUstream
    ) nogil

    cdef CUresult cccl_device_reduce_cleanup(
        cccl_device_reduce_build_result_t*
    ) nogil


cdef class DeviceReduceBuildResult:
    cdef cccl_device_reduce_build_result_t build_data

    def __cinit__(
        DeviceReduceBuildResult self,
        Iterator d_in,
        Iterator d_out,
        Op op,
        Value h_init,
        cccl_determinism_t determinism,
        CommonData common_data
    ):
        cdef CUresult status = -1
        cdef int cc_major = common_data.get_cc_major()
        cdef int cc_minor = common_data.get_cc_minor()
        cdef const char *cub_path = common_data.cub_path_get_c_str()
        cdef const char *thrust_path = common_data.thrust_path_get_c_str()
        cdef const char *libcudacxx_path = common_data.libcudacxx_path_get_c_str()
        cdef const char *ctk_path = common_data.ctk_path_get_c_str()
        memset(&self.build_data, 0, sizeof(cccl_device_reduce_build_result_t))

        with nogil:
            status = cccl_device_reduce_build(
                &self.build_data,
                d_in.iter_data,
                d_out.iter_data,
                op.op_data,
                h_init.value_data,
                determinism,
                cc_major,
                cc_minor,
                cub_path,
                thrust_path,
                libcudacxx_path,
                ctk_path,
            )
        if status != 0:
            raise RuntimeError(
                f"Failed building reduce, error code: {status}"
            )

    def __dealloc__(DeviceReduceBuildResult self):
        cdef CUresult status = -1
        with nogil:
            status = cccl_device_reduce_cleanup(&self.build_data)
        if (status != 0):
            print(f"Return code {status} encountered during reduce result cleanup")

    cpdef size_t compute(
        DeviceReduceBuildResult self,
        temp_storage_ptr,
        temp_storage_bytes,
        Iterator d_in,
        Iterator d_out,
        size_t num_items,
        Op op,
        Value h_init,
        stream
    ):
        cdef CUresult status = -1
        cdef void *storage_ptr = (<void *><uintptr_t>temp_storage_ptr) if temp_storage_ptr else NULL
        cdef size_t storage_sz = <size_t>temp_storage_bytes
        cdef CUstream c_stream = <CUstream><uintptr_t>(stream) if stream else NULL

        with nogil:
            status = cccl_device_reduce(
                self.build_data,
                storage_ptr,
                &storage_sz,
                d_in.iter_data,
                d_out.iter_data,
                <uint64_t>num_items,
                op.op_data,
                h_init.value_data,
                c_stream
            )
        if status != 0:
            raise RuntimeError(
                f"Failed executing reduce, error code: {status}"
            )
        return storage_sz

    cpdef size_t compute_nondeterministic(
        DeviceReduceBuildResult self,
        temp_storage_ptr,
        temp_storage_bytes,
        Iterator d_in,
        Iterator d_out,
        size_t num_items,
        Op op,
        Value h_init,
        stream
    ):
        cdef CUresult status = -1
        cdef void *storage_ptr = (<void *><uintptr_t>temp_storage_ptr) if temp_storage_ptr else NULL
        cdef size_t storage_sz = <size_t>temp_storage_bytes
        cdef CUstream c_stream = <CUstream><uintptr_t>(stream) if stream else NULL

        with nogil:
            status = cccl_device_reduce_nondeterministic(
                self.build_data,
                storage_ptr,
                &storage_sz,
                d_in.iter_data,
                d_out.iter_data,
                <uint64_t>num_items,
                op.op_data,
                h_init.value_data,
                c_stream
            )
        if status != 0:
            raise RuntimeError(
                f"Failed executing reduce not guaranteed determinism, error code: {status}"
            )
        return storage_sz

    def _get_cubin(self):
        return PyBytes_FromStringAndSize(
            <const char*>self.build_data.cubin,
            self.build_data.cubin_size
        )

# ------------
#   DeviceScan
# ------------


cdef extern from "cccl/c/scan.h":
    ctypedef bint _Bool

    cdef struct cccl_device_scan_build_result_t 'cccl_device_scan_build_result_t':
        const char* cubin
        size_t cubin_size

    cdef CUresult cccl_device_scan_build(
        cccl_device_scan_build_result_t*,
        cccl_iterator_t,
        cccl_iterator_t,
        cccl_op_t,
        cccl_type_info,
        _Bool,
        cccl_init_kind_t,
        int, int, const char*, const char*, const char*, const char*
    ) nogil

    cdef CUresult cccl_device_exclusive_scan(
        cccl_device_scan_build_result_t,
        void *,
        size_t *,
        cccl_iterator_t,
        cccl_iterator_t,
        uint64_t,
        cccl_op_t,
        cccl_value_t,
        CUstream
    ) nogil

    cdef CUresult cccl_device_inclusive_scan(
        cccl_device_scan_build_result_t,
        void *,
        size_t *,
        cccl_iterator_t,
        cccl_iterator_t,
        uint64_t,
        cccl_op_t,
        cccl_value_t,
        CUstream
    ) nogil

    cdef CUresult cccl_device_exclusive_scan_future_value(
        cccl_device_scan_build_result_t,
        void *,
        size_t *,
        cccl_iterator_t,
        cccl_iterator_t,
        uint64_t,
        cccl_op_t,
        cccl_iterator_t,
        CUstream
    ) nogil

    cdef CUresult cccl_device_inclusive_scan_future_value(
        cccl_device_scan_build_result_t,
        void *,
        size_t *,
        cccl_iterator_t,
        cccl_iterator_t,
        uint64_t,
        cccl_op_t,
        cccl_iterator_t,
        CUstream
    ) nogil

    cdef CUresult cccl_device_inclusive_scan_no_init(
        cccl_device_scan_build_result_t,
        void *,
        size_t *,
        cccl_iterator_t,
        cccl_iterator_t,
        uint64_t,
        cccl_op_t,
        CUstream
    ) nogil

    cdef CUresult cccl_device_scan_cleanup(
        cccl_device_scan_build_result_t*
    ) nogil


cdef class DeviceScanBuildResult:
    cdef cccl_device_scan_build_result_t build_data

    def __cinit__(
        DeviceScanBuildResult self,
        Iterator d_in,
        Iterator d_out,
        Op op,
        TypeInfo init_type,
        bint force_inclusive,
        cccl_init_kind_t init_kind,
        CommonData common_data
    ):
        cdef CUresult status = -1
        cdef int cc_major = common_data.get_cc_major()
        cdef int cc_minor = common_data.get_cc_minor()
        cdef const char *cub_path = common_data.cub_path_get_c_str()
        cdef const char *thrust_path = common_data.thrust_path_get_c_str()
        cdef const char *libcudacxx_path = common_data.libcudacxx_path_get_c_str()
        cdef const char *ctk_path = common_data.ctk_path_get_c_str()
        memset(&self.build_data, 0, sizeof(cccl_device_scan_build_result_t))

        with nogil:
            status = cccl_device_scan_build(
                &self.build_data,
                d_in.iter_data,
                d_out.iter_data,
                op.op_data,
                init_type.type_info,
                force_inclusive,
                init_kind,
                cc_major,
                cc_minor,
                cub_path,
                thrust_path,
                libcudacxx_path,
                ctk_path,
            )
        if status != 0:
            raise RuntimeError(f"Error {status} building scan")

    def __dealloc__(DeviceScanBuildResult self):
        cdef CUresult status = -1
        with nogil:
            status = cccl_device_scan_cleanup(&self.build_data)
        if (status != 0):
            print(f"Return code {status} encountered during scan result cleanup")

    cpdef size_t compute_inclusive(
        DeviceScanBuildResult self,
        temp_storage_ptr,
        temp_storage_bytes,
        Iterator d_in,
        Iterator d_out,
        size_t num_items,
        Op op,
        Value init_value,
        stream
    ):
        cdef CUresult status = -1
        cdef void *storage_ptr = (<void *><uintptr_t>temp_storage_ptr) if temp_storage_ptr else NULL
        cdef size_t storage_sz = <size_t>temp_storage_bytes
        cdef CUstream c_stream = <CUstream><uintptr_t>(stream) if stream else NULL

        with nogil:
            status = cccl_device_inclusive_scan(
                self.build_data,
                storage_ptr,
                &storage_sz,
                d_in.iter_data,
                d_out.iter_data,
                <uint64_t>num_items,
                op.op_data,
                init_value.value_data,
                c_stream
            )
        if status != 0:
            raise RuntimeError(
                f"Failed executing inclusive scan, error code: {status}"
            )
        return storage_sz

    cpdef size_t compute_exclusive(
        DeviceScanBuildResult self,
        temp_storage_ptr,
        temp_storage_bytes,
        Iterator d_in,
        Iterator d_out,
        size_t num_items,
        Op op,
        Value init_value,
        stream
    ):
        cdef CUresult status = -1
        cdef void *storage_ptr = (<void *><uintptr_t>temp_storage_ptr) if temp_storage_ptr else NULL
        cdef size_t storage_sz = <size_t>temp_storage_bytes
        cdef CUstream c_stream = <CUstream><uintptr_t>(stream) if stream else NULL

        with nogil:
            status = cccl_device_exclusive_scan(
                self.build_data,
                storage_ptr,
                &storage_sz,
                d_in.iter_data,
                d_out.iter_data,
                <uint64_t>num_items,
                op.op_data,
                init_value.value_data,
                c_stream
            )
        if status != 0:
            raise RuntimeError(
                f"Failed executing exclusive scan, error code: {status}"
            )
        return storage_sz

    cpdef size_t compute_inclusive_future_value(
        DeviceScanBuildResult self,
        temp_storage_ptr,
        temp_storage_bytes,
        Iterator d_in,
        Iterator d_out,
        size_t num_items,
        Op op,
        Iterator init_value,
        stream
    ):
        cdef CUresult status = -1
        cdef void *storage_ptr = (<void *><uintptr_t>temp_storage_ptr) if temp_storage_ptr else NULL
        cdef size_t storage_sz = <size_t>temp_storage_bytes
        cdef CUstream c_stream = <CUstream><uintptr_t>(stream) if stream else NULL

        with nogil:
            status = cccl_device_inclusive_scan_future_value(
                self.build_data,
                storage_ptr,
                &storage_sz,
                d_in.iter_data,
                d_out.iter_data,
                <uint64_t>num_items,
                op.op_data,
                init_value.iter_data,
                c_stream
            )
        if status != 0:
            raise RuntimeError(
                f"Failed executing inclusive scan, error code: {status}"
            )
        return storage_sz

    cpdef size_t compute_exclusive_future_value(
        DeviceScanBuildResult self,
        temp_storage_ptr,
        temp_storage_bytes,
        Iterator d_in,
        Iterator d_out,
        size_t num_items,
        Op op,
        Iterator init_value,
        stream
    ):
        cdef CUresult status = -1
        cdef void *storage_ptr = (<void *><uintptr_t>temp_storage_ptr) if temp_storage_ptr else NULL
        cdef size_t storage_sz = <size_t>temp_storage_bytes
        cdef CUstream c_stream = <CUstream><uintptr_t>(stream) if stream else NULL

        with nogil:
            status = cccl_device_exclusive_scan_future_value(
                self.build_data,
                storage_ptr,
                &storage_sz,
                d_in.iter_data,
                d_out.iter_data,
                <uint64_t>num_items,
                op.op_data,
                init_value.iter_data,
                c_stream
            )
        if status != 0:
            raise RuntimeError(
                f"Failed executing exclusive scan, error code: {status}"
            )
        return storage_sz

    cpdef size_t compute_inclusive_no_init(
        DeviceScanBuildResult self,
        temp_storage_ptr,
        temp_storage_bytes,
        Iterator d_in,
        Iterator d_out,
        size_t num_items,
        Op op,
        object init_value,
        stream
    ):
        cdef CUresult status = -1
        cdef void *storage_ptr = (<void *><uintptr_t>temp_storage_ptr) if temp_storage_ptr else NULL
        cdef size_t storage_sz = <size_t>temp_storage_bytes
        cdef CUstream c_stream = <CUstream><uintptr_t>(stream) if stream else NULL

        with nogil:
            status = cccl_device_inclusive_scan_no_init(
                self.build_data,
                storage_ptr,
                &storage_sz,
                d_in.iter_data,
                d_out.iter_data,
                <uint64_t>num_items,
                op.op_data,
                c_stream
            )
        if status != 0:
            raise RuntimeError(
                f"Failed executing inclusive scan, error code: {status}"
            )
        return storage_sz

    def _get_cubin(self):
        return PyBytes_FromStringAndSize(
            <const char*>self.build_data.cubin,
            self.build_data.cubin_size
        )

# -----------------------
#   DeviceSegmentedReduce
# -----------------------


cdef extern from "cccl/c/segmented_reduce.h":
    cdef struct cccl_device_segmented_reduce_build_result_t 'cccl_device_segmented_reduce_build_result_t':
        const char* cubin
        size_t cubin_size

    cdef CUresult cccl_device_segmented_reduce_build(
        cccl_device_segmented_reduce_build_result_t*,
        cccl_iterator_t,
        cccl_iterator_t,
        cccl_iterator_t,
        cccl_iterator_t,
        cccl_op_t,
        cccl_value_t,
        int, int, const char*, const char*, const char*, const char*
    ) nogil

    cdef CUresult cccl_device_segmented_reduce(
        cccl_device_segmented_reduce_build_result_t,
        void *,
        size_t *,
        cccl_iterator_t,
        cccl_iterator_t,
        uint64_t,
        cccl_iterator_t,
        cccl_iterator_t,
        cccl_op_t,
        cccl_value_t,
        size_t,
        CUstream
    ) nogil

    cdef CUresult cccl_device_segmented_reduce_cleanup(
        cccl_device_segmented_reduce_build_result_t* bld_ptr
    ) nogil


cdef class DeviceSegmentedReduceBuildResult:
    cdef cccl_device_segmented_reduce_build_result_t build_data

    def __cinit__(
        DeviceSegmentedReduceBuildResult self,
        Iterator d_in,
        Iterator d_out,
        Iterator start_offsets,
        Iterator end_offsets,
        Op op,
        Value h_init,
        CommonData common_data
    ):
        cdef CUresult status = -1
        cdef int cc_major = common_data.get_cc_major()
        cdef int cc_minor = common_data.get_cc_minor()
        cdef const char *cub_path = common_data.cub_path_get_c_str()
        cdef const char *thrust_path = common_data.thrust_path_get_c_str()
        cdef const char *libcudacxx_path = common_data.libcudacxx_path_get_c_str()
        cdef const char *ctk_path = common_data.ctk_path_get_c_str()

        memset(&self.build_data, 0, sizeof(cccl_device_segmented_reduce_build_result_t))
        with nogil:
            status = cccl_device_segmented_reduce_build(
                &self.build_data,
                d_in.iter_data,
                d_out.iter_data,
                start_offsets.iter_data,
                end_offsets.iter_data,
                op.op_data,
                h_init.value_data,
                cc_major,
                cc_minor,
                cub_path,
                thrust_path,
                libcudacxx_path,
                ctk_path,
            )
        if status != 0:
            raise RuntimeError(
                f"Failed building segmented_reduce, error code: {status}"
            )

    def __dealloc__(DeviceSegmentedReduceBuildResult self):
        cdef CUresult status = -1
        with nogil:
            status = cccl_device_segmented_reduce_cleanup(&self.build_data)
        if (status != 0):
            print(f"Return code {status} encountered during segmented_reduce result cleanup")

    cpdef size_t compute(
        DeviceSegmentedReduceBuildResult self,
        temp_storage_ptr,
        temp_storage_bytes,
        Iterator d_in,
        Iterator d_out,
        size_t num_items,
        Iterator start_offsets,
        Iterator end_offsets,
        Op op,
        Value h_init,
        size_t max_segment_size=0,
        stream=None
    ):
        cdef CUresult status = -1
        cdef void *storage_ptr = (<void *><uintptr_t>temp_storage_ptr) if temp_storage_ptr else NULL
        cdef size_t storage_sz = <size_t>temp_storage_bytes
        cdef CUstream c_stream = <CUstream><uintptr_t>(stream) if stream else NULL

        with nogil:
            status = cccl_device_segmented_reduce(
                self.build_data,
                storage_ptr,
                &storage_sz,
                d_in.iter_data,
                d_out.iter_data,
                <uint64_t>num_items,
                start_offsets.iter_data,
                end_offsets.iter_data,
                op.op_data,
                h_init.value_data,
                max_segment_size,
                c_stream
            )
        if status != 0:
            raise RuntimeError(
                f"Failed executing segmented_reduce, error code: {status}"
            )
        return storage_sz

    def _get_cubin(self):
        return PyBytes_FromStringAndSize(
            <const char*>self.build_data.cubin,
            self.build_data.cubin_size
        )

# -----------------
#   DeviceMergeSort
# -----------------


cdef extern from "cccl/c/merge_sort.h":
    cdef struct cccl_device_merge_sort_build_result_t 'cccl_device_merge_sort_build_result_t':
        const char* cubin
        size_t cubin_size

    cdef CUresult cccl_device_merge_sort_build(
        cccl_device_merge_sort_build_result_t *bld_ptr,
        cccl_iterator_t d_in_keys,
        cccl_iterator_t d_in_items,
        cccl_iterator_t d_out_keys,
        cccl_iterator_t d_out_items,
        cccl_op_t,
        int, int, const char*, const char*, const char*, const char*
    ) nogil

    cdef CUresult cccl_device_merge_sort(
        cccl_device_merge_sort_build_result_t,
        void *,
        size_t *,
        cccl_iterator_t,
        cccl_iterator_t,
        cccl_iterator_t,
        cccl_iterator_t,
        uint64_t,
        cccl_op_t,
        CUstream
    ) nogil

    cdef CUresult cccl_device_merge_sort_cleanup(
        cccl_device_merge_sort_build_result_t* bld_ptr
    ) nogil


cdef class DeviceMergeSortBuildResult:
    cdef cccl_device_merge_sort_build_result_t build_data

    def __cinit__(
        DeviceMergeSortBuildResult self,
        Iterator d_in_keys,
        Iterator d_in_items,
        Iterator d_out_keys,
        Iterator d_out_items,
        Op op,
        CommonData common_data
    ):
        cdef CUresult status = -1
        cdef int cc_major = common_data.get_cc_major()
        cdef int cc_minor = common_data.get_cc_minor()
        cdef const char *cub_path = common_data.cub_path_get_c_str()
        cdef const char *thrust_path = common_data.thrust_path_get_c_str()
        cdef const char *libcudacxx_path = common_data.libcudacxx_path_get_c_str()
        cdef const char *ctk_path = common_data.ctk_path_get_c_str()

        memset(&self.build_data, 0, sizeof(cccl_device_merge_sort_build_result_t))
        with nogil:
            status = cccl_device_merge_sort_build(
                &self.build_data,
                d_in_keys.iter_data,
                d_in_items.iter_data,
                d_out_keys.iter_data,
                d_out_items.iter_data,
                op.op_data,
                cc_major,
                cc_minor,
                cub_path,
                thrust_path,
                libcudacxx_path,
                ctk_path,
            )
        if status != 0:
            raise RuntimeError(
                f"Failed building merge_sort, error code: {status}"
            )

    def __dealloc__(DeviceMergeSortBuildResult self):
        cdef CUresult status = -1
        with nogil:
            status = cccl_device_merge_sort_cleanup(&self.build_data)
        if (status != 0):
            print(f"Return code {status} encountered during merge_sort result cleanup")

    cpdef size_t compute(
        DeviceMergeSortBuildResult self,
        temp_storage_ptr,
        temp_storage_bytes,
        Iterator d_in_keys,
        Iterator d_in_items,
        Iterator d_out_keys,
        Iterator d_out_items,
        size_t num_items,
        Op op,
        stream
    ):
        cdef CUresult status = -1
        cdef void *storage_ptr = (<void *><uintptr_t>temp_storage_ptr) if temp_storage_ptr else NULL
        cdef size_t storage_sz = <size_t>temp_storage_bytes
        cdef CUstream c_stream = <CUstream><uintptr_t>(stream) if stream else NULL
        with nogil:
            status = cccl_device_merge_sort(
                self.build_data,
                storage_ptr,
                &storage_sz,
                d_in_keys.iter_data,
                d_in_items.iter_data,
                d_out_keys.iter_data,
                d_out_items.iter_data,
                <uint64_t>num_items,
                op.op_data,
                c_stream
            )
        if status != 0:
            raise RuntimeError(
                f"Failed executing merge_sort, error code: {status}"
            )
        return storage_sz


    def _get_cubin(self):
        return PyBytes_FromStringAndSize(
            <const char*>self.build_data.cubin,
            self.build_data.cubin_size
        )


# -------------------
#   DeviceUniqueByKey
# -------------------

cdef extern from "cccl/c/unique_by_key.h":
    cdef struct cccl_device_unique_by_key_build_result_t 'cccl_device_unique_by_key_build_result_t':
        const char* cubin
        size_t cubin_size


    cdef CUresult cccl_device_unique_by_key_build(
        cccl_device_unique_by_key_build_result_t *build_ptr,
        cccl_iterator_t d_keys_in,
        cccl_iterator_t d_values_in,
        cccl_iterator_t d_keys_out,
        cccl_iterator_t d_values_out,
        cccl_iterator_t d_num_selected_out,
        cccl_op_t comparison_op,
        int, int, const char *, const char *, const char *, const char *
    ) nogil

    cdef CUresult cccl_device_unique_by_key(
        cccl_device_unique_by_key_build_result_t build,
        void *d_storage_ptr,
        size_t *d_storage_nbytes,
        cccl_iterator_t d_keys_in,
        cccl_iterator_t d_values_in,
        cccl_iterator_t d_keys_out,
        cccl_iterator_t d_values_out,
        cccl_iterator_t d_num_selected_out,
        cccl_op_t comparison_op,
        size_t num_items,
        CUstream stream
    ) nogil

    cdef CUresult cccl_device_unique_by_key_cleanup(
        cccl_device_unique_by_key_build_result_t *build_ptr,
    ) nogil


cdef class DeviceUniqueByKeyBuildResult:
    cdef cccl_device_unique_by_key_build_result_t build_data

    def __cinit__(
        DeviceUniqueByKeyBuildResult self,
        Iterator d_keys_in,
        Iterator d_values_in,
        Iterator d_keys_out,
        Iterator d_values_out,
        Iterator d_num_selected_out,
        Op comparison_op,
        CommonData common_data
    ):
        cdef CUresult status = -1
        cdef int cc_major = common_data.get_cc_major()
        cdef int cc_minor = common_data.get_cc_minor()
        cdef const char *cub_path = common_data.cub_path_get_c_str()
        cdef const char *thrust_path = common_data.thrust_path_get_c_str()
        cdef const char *libcudacxx_path = common_data.libcudacxx_path_get_c_str()
        cdef const char *ctk_path = common_data.ctk_path_get_c_str()

        memset(&self.build_data, 0, sizeof(cccl_device_unique_by_key_build_result_t))
        with nogil:
            status = cccl_device_unique_by_key_build(
                &self.build_data,
                d_keys_in.iter_data,
                d_values_in.iter_data,
                d_keys_out.iter_data,
                d_values_out.iter_data,
                d_num_selected_out.iter_data,
                comparison_op.op_data,
                cc_major,
                cc_minor,
                cub_path,
                thrust_path,
                libcudacxx_path,
                ctk_path,
            )
        if status != 0:
            raise RuntimeError(
                f"Failed building unique_by_key, error code: {status}"
            )

    def __dealloc__(DeviceUniqueByKeyBuildResult self):
        cdef CUresult status = -1
        with nogil:
            status = cccl_device_unique_by_key_cleanup(&self.build_data)
        if (status != 0):
            print(f"Return code {status} encountered during unique_by_key result cleanup")

    cpdef size_t compute(
        DeviceUniqueByKeyBuildResult self,
        temp_storage_ptr,
        temp_storage_bytes,
        Iterator d_keys_in,
        Iterator d_values_in,
        Iterator d_keys_out,
        Iterator d_values_out,
        Iterator d_num_selected_out,
        Op comparison_op,
        size_t num_items,
        stream
    ):
        cdef CUresult status = -1
        cdef void *storage_ptr = (<void *><uintptr_t>temp_storage_ptr) if temp_storage_ptr else NULL
        cdef size_t storage_sz = <size_t>temp_storage_bytes
        cdef CUstream c_stream = <CUstream><uintptr_t>(stream) if stream else NULL

        with nogil:
            status = cccl_device_unique_by_key(
                self.build_data,
                storage_ptr,
                &storage_sz,
                d_keys_in.iter_data,
                d_values_in.iter_data,
                d_keys_out.iter_data,
                d_values_out.iter_data,
                d_num_selected_out.iter_data,
                comparison_op.op_data,
                <uint64_t>num_items,
                c_stream
            )

        if status != 0:
            raise RuntimeError(
                f"Failed executing unique_by_key, error code: {status}"
            )
        return storage_sz

    def _get_cubin(self):
        return PyBytes_FromStringAndSize(
            <const char*>self.build_data.cubin,
            self.build_data.cubin_size
        )

# -----------------
# DeviceRadixSort
# -----------------

cdef extern from "cccl/c/radix_sort.h":
    cdef struct cccl_device_radix_sort_build_result_t 'cccl_device_radix_sort_build_result_t':
        const char* cubin
        size_t cubin_size

    cdef CUresult cccl_device_radix_sort_build(
        cccl_device_radix_sort_build_result_t *build_ptr,
        cccl_sort_order_t sort_order,
        cccl_iterator_t d_keys_in,
        cccl_iterator_t d_values_in,
        cccl_op_t decomposer,
        const char* decomposer_return_type,
        int, int, const char *, const char *, const char *, const char *
    ) nogil

    cdef CUresult cccl_device_radix_sort(
        cccl_device_radix_sort_build_result_t build,
        void *d_storage_ptr,
        size_t *d_storage_nbytes,
        cccl_iterator_t d_keys_in,
        cccl_iterator_t d_keys_out,
        cccl_iterator_t d_values_in,
        cccl_iterator_t d_values_out,
        cccl_op_t decomposer,
        size_t num_items,
        int begin_bit,
        int end_bit,
        bint is_overwrite_okay,
        int* selector,
        CUstream stream
    ) nogil

    cdef CUresult cccl_device_radix_sort_cleanup(
        cccl_device_radix_sort_build_result_t *build_ptr,
    ) nogil


cdef class DeviceRadixSortBuildResult:
    cdef cccl_device_radix_sort_build_result_t build_data

    def __dealloc__(DeviceRadixSortBuildResult self):
        cdef CUresult status = -1
        with nogil:
            status = cccl_device_radix_sort_cleanup(&self.build_data)
        if (status != 0):
            print(f"Return code {status} encountered during radix_sort result cleanup")

    def __cinit__(
        DeviceRadixSortBuildResult self,
        cccl_sort_order_t order,
        Iterator d_keys_in,
        Iterator d_values_in,
        Op decomposer_op,
        const char* decomposer_return_type,
        CommonData common_data
    ):
        cdef CUresult status = -1
        cdef int cc_major = common_data.get_cc_major()
        cdef int cc_minor = common_data.get_cc_minor()
        cdef const char *cub_path = common_data.cub_path_get_c_str()
        cdef const char *thrust_path = common_data.thrust_path_get_c_str()
        cdef const char *libcudacxx_path = common_data.libcudacxx_path_get_c_str()
        cdef const char *ctk_path = common_data.ctk_path_get_c_str()

        memset(&self.build_data, 0, sizeof(cccl_device_radix_sort_build_result_t))
        with nogil:
            status = cccl_device_radix_sort_build(
                &self.build_data,
                order,
                d_keys_in.iter_data,
                d_values_in.iter_data,
                decomposer_op.op_data,
                decomposer_return_type,
                cc_major,
                cc_minor,
                cub_path,
                thrust_path,
                libcudacxx_path,
                ctk_path,
            )
        if status != 0:
            raise RuntimeError(
                f"Failed building radix_sort, error code: {status}"
            )

    cpdef tuple compute(
        DeviceRadixSortBuildResult self,
        temp_storage_ptr,
        temp_storage_bytes,
        Iterator d_keys_in,
        Iterator d_keys_out,
        Iterator d_values_in,
        Iterator d_values_out,
        Op decomposer_op,
        size_t num_items,
        int begin_bit,
        int end_bit,
        bint is_overwrite_okay,
        selector,
        stream
    ):
        cdef CUresult status = -1
        cdef void *storage_ptr = (<void *><size_t>temp_storage_ptr) if temp_storage_ptr else NULL
        cdef size_t storage_sz = <size_t>temp_storage_bytes
        cdef int selector_int = <int>selector
        cdef CUstream c_stream = <CUstream><size_t>(stream) if stream else NULL

        with nogil:
            status = cccl_device_radix_sort(
                self.build_data,
                storage_ptr,
                &storage_sz,
                d_keys_in.iter_data,
                d_keys_out.iter_data,
                d_values_in.iter_data,
                d_values_out.iter_data,
                decomposer_op.op_data,
                <uint64_t>num_items,
                begin_bit,
                end_bit,
                is_overwrite_okay,
                &selector_int,
                c_stream
            )

        if status != 0:
            raise RuntimeError(
                f"Failed executing radix_sort, error code: {status}"
            )
        return <object>storage_sz, <object>selector_int


    def _get_cubin(self):
        return PyBytes_FromStringAndSize(
            <const char*>self.build_data.cubin,
            self.build_data.cubin_size
        )

# --------------------------------------------
#   DeviceUnaryTransform/DeviceBinaryTransform
# --------------------------------------------
cdef extern from "cccl/c/transform.h":
    cdef struct cccl_device_transform_build_result_t:
        const char* cubin
        size_t cubin_size

    cdef CUresult cccl_device_unary_transform_build(
        cccl_device_transform_build_result_t *build_ptr,
        cccl_iterator_t d_in,
        cccl_iterator_t d_out,
        cccl_op_t op,
        int, int, const char *, const char *, const char *, const char *
    ) nogil

    cdef CUresult cccl_device_unary_transform(
      cccl_device_transform_build_result_t build,
      cccl_iterator_t d_in,
      cccl_iterator_t d_out,
      uint64_t num_items,
      cccl_op_t op,
      CUstream stream) nogil

    cdef CUresult cccl_device_binary_transform_build(
      cccl_device_transform_build_result_t* build_ptr,
      cccl_iterator_t d_in1,
      cccl_iterator_t d_in2,
      cccl_iterator_t d_out,
      cccl_op_t op,
      int, int, const char *, const char *, const char *, const char *
    ) nogil

    cdef CUresult cccl_device_binary_transform(
      cccl_device_transform_build_result_t build,
      cccl_iterator_t d_in1,
      cccl_iterator_t d_in2,
      cccl_iterator_t d_out,
      uint64_t num_items,
      cccl_op_t op,
      CUstream stream) nogil

    cdef CUresult cccl_device_transform_cleanup(
        cccl_device_transform_build_result_t *build_ptr,
    ) nogil


cdef class DeviceUnaryTransform:
    cdef cccl_device_transform_build_result_t build_data

    def __cinit__(
        self,
        Iterator d_in,
        Iterator d_out,
        Op op,
        CommonData common_data
    ):
        memset(&self.build_data, 0, sizeof(cccl_device_transform_build_result_t))

        cdef CUresult status = -1
        cdef int cc_major = common_data.get_cc_major()
        cdef int cc_minor = common_data.get_cc_minor()
        cdef const char *cub_path = common_data.cub_path_get_c_str()
        cdef const char *thrust_path = common_data.thrust_path_get_c_str()
        cdef const char *libcudacxx_path = common_data.libcudacxx_path_get_c_str()
        cdef const char *ctk_path = common_data.ctk_path_get_c_str()

        with nogil:
            status = cccl_device_unary_transform_build(
                &self.build_data,
                d_in.iter_data,
                d_out.iter_data,
                op.op_data,
                cc_major,
                cc_minor,
                cub_path,
                thrust_path,
                libcudacxx_path,
                ctk_path,
            )
        if status != 0:
            raise RuntimeError("Failed to build unary transform")

    def __dealloc__(DeviceUnaryTransform self):
        cdef CUresult status = -1
        with nogil:
            status = cccl_device_transform_cleanup(&self.build_data)
        if (status != 0):
            print(f"Return code {status} encountered during unary transform result cleanup")

    cpdef void compute(
        DeviceUnaryTransform self,
        Iterator d_in,
        Iterator d_out,
        size_t num_items,
        Op op,
        stream
    ):
        cdef CUresult status = -1
        cdef CUstream c_stream = <CUstream><uintptr_t>(stream) if stream else NULL
        with nogil:
            status = cccl_device_unary_transform(
                self.build_data,
                d_in.iter_data,
                d_out.iter_data,
                <uint64_t>num_items,
                op.op_data,
                c_stream
            )
        if (status != 0):
            raise RuntimeError("Failed to compute unary transform")


    def _get_cubin(self):
        return PyBytes_FromStringAndSize(
            <const char*>self.build_data.cubin,
            self.build_data.cubin_size
        )


cdef class DeviceBinaryTransform:
    cdef cccl_device_transform_build_result_t build_data

    def __cinit__(
        self,
        Iterator d_in1,
        Iterator d_in2,
        Iterator d_out,
        Op op,
        CommonData common_data
    ):
        memset(&self.build_data, 0, sizeof(cccl_device_transform_build_result_t))

        cdef CUresult status = -1
        cdef int cc_major = common_data.get_cc_major()
        cdef int cc_minor = common_data.get_cc_minor()
        cdef const char *cub_path = common_data.cub_path_get_c_str()
        cdef const char *thrust_path = common_data.thrust_path_get_c_str()
        cdef const char *libcudacxx_path = common_data.libcudacxx_path_get_c_str()
        cdef const char *ctk_path = common_data.ctk_path_get_c_str()

        with nogil:
            status = cccl_device_binary_transform_build(
                &self.build_data,
                d_in1.iter_data,
                d_in2.iter_data,
                d_out.iter_data,
                op.op_data,
                cc_major,
                cc_minor,
                cub_path,
                thrust_path,
                libcudacxx_path,
                ctk_path,
            )
        if status != 0:
            raise RuntimeError("Failed to build binary transform")

    def __dealloc__(DeviceBinaryTransform self):
        cdef CUresult status = -1
        with nogil:
            status = cccl_device_transform_cleanup(&self.build_data)
        if (status != 0):
            print(f"Return code {status} encountered during binary transform result cleanup")

    cpdef void compute(
        DeviceBinaryTransform self,
        Iterator d_in1,
        Iterator d_in2,
        Iterator d_out,
        size_t num_items,
        Op op,
        stream
    ):
        cdef CUresult status = -1
        cdef CUstream c_stream = <CUstream><uintptr_t>(stream) if stream else NULL
        with nogil:
            status = cccl_device_binary_transform(
                self.build_data,
                d_in1.iter_data,
                d_in2.iter_data,
                d_out.iter_data,
                <uint64_t>num_items,
                op.op_data,
                c_stream
            )
        if (status != 0):
            raise RuntimeError("Failed to compute binary transform")

    def _get_cubin(self):
        return PyBytes_FromStringAndSize(
            <const char*>self.build_data.cubin,
            self.build_data.cubin_size
        )


# -----------------
#   DeviceHistogram
# -----------------
cdef extern from "cccl/c/histogram.h":
    cdef struct cccl_device_histogram_build_result_t 'cccl_device_histogram_build_result_t':
        const char* cubin
        size_t cubin_size

    cdef CUresult cccl_device_histogram_build(
        cccl_device_histogram_build_result_t *build_ptr,
        int num_channels,
        int num_active_channels,
        cccl_iterator_t d_samples,
        int num_output_levels_val,
        cccl_iterator_t d_output_histograms,
        cccl_value_t h_levels,
        int64_t num_rows,
        int64_t row_stride_samples,
        bint is_evenly_segmented,
        int, int, const char *, const char *, const char *, const char *
    ) nogil

    cdef CUresult cccl_device_histogram_even(
        cccl_device_histogram_build_result_t build,
        void *d_storage_ptr,
        size_t *d_storage_nbytes,
        cccl_iterator_t d_samples,
        cccl_iterator_t d_output_histograms,
        cccl_value_t num_output_levels,
        cccl_value_t lower_level,
        cccl_value_t upper_level,
        int64_t num_row_pixels,
        int64_t num_rows,
        int64_t row_stride_samples,
        CUstream stream
    ) nogil

    cdef CUresult cccl_device_histogram_cleanup(
        cccl_device_histogram_build_result_t *build_ptr,
    ) nogil


cdef class DeviceHistogramBuildResult:
    cdef cccl_device_histogram_build_result_t build_data

    def __dealloc__(DeviceHistogramBuildResult self):
        cdef CUresult status = -1
        with nogil:
            status = cccl_device_histogram_cleanup(&self.build_data)
        if (status != 0):
            print(f"Return code {status} encountered during histogram result cleanup")


    def __cinit__(
        DeviceHistogramBuildResult self,
        int num_channels,
        int num_active_channels,
        Iterator d_samples,
        int num_levels,
        Iterator d_histogram,
        Value h_levels,
        int num_rows,
        int row_stride_samples,
        bint is_evenly_segmented,
        CommonData common_data
    ):
        cdef CUresult status = -1
        cdef int cc_major = common_data.get_cc_major()
        cdef int cc_minor = common_data.get_cc_minor()
        cdef const char *cub_path = common_data.cub_path_get_c_str()
        cdef const char *thrust_path = common_data.thrust_path_get_c_str()
        cdef const char *libcudacxx_path = common_data.libcudacxx_path_get_c_str()
        cdef const char *ctk_path = common_data.ctk_path_get_c_str()

        memset(&self.build_data, 0, sizeof(cccl_device_histogram_build_result_t))
        with nogil:
            status = cccl_device_histogram_build(
                &self.build_data,
                num_channels,
                num_active_channels,
                d_samples.iter_data,
                num_levels,
                d_histogram.iter_data,
                h_levels.value_data,
                num_rows,
                row_stride_samples,
                is_evenly_segmented,
                cc_major,
                cc_minor,
                cub_path,
                thrust_path,
                libcudacxx_path,
                ctk_path,
            )
        if status != 0:
            raise RuntimeError(
                f"Failed building histogram, error code: {status}"
            )

    cpdef size_t compute_even(
        DeviceHistogramBuildResult self,
        temp_storage_ptr,
        temp_storage_bytes,
        Iterator d_samples,
        Iterator d_histogram,
        Value h_num_output_levels,
        Value h_lower_level,
        Value h_upper_level,
        int num_row_pixels,
        int num_rows,
        int row_stride_samples,
        stream
    ):
        cdef CUresult status = -1
        cdef void *storage_ptr = (<void *><size_t>temp_storage_ptr) if temp_storage_ptr else NULL
        cdef size_t storage_sz = <size_t>temp_storage_bytes
        cdef CUstream c_stream = <CUstream><size_t>(stream) if stream else NULL

        with nogil:
            status = cccl_device_histogram_even(
                self.build_data,
                storage_ptr,
                &storage_sz,
                d_samples.iter_data,
                d_histogram.iter_data,
                h_num_output_levels.value_data,
                h_lower_level.value_data,
                h_upper_level.value_data,
                num_row_pixels,
                num_rows,
                row_stride_samples,
                c_stream
            )
        if status != 0:
            raise RuntimeError(
                f"Failed executing histogram, error code: {status}"
            )
        return storage_sz


    def _get_cubin(self):
        return PyBytes_FromStringAndSize(
            <const char*>self.build_data.cubin,
            self.build_data.cubin_size
        )


# -------------------
#   DeviceBinarySearch
# -------------------
cdef extern from "cccl/c/binary_search.h":
    cdef struct cccl_device_binary_search_build_result_t 'cccl_device_binary_search_build_result_t':
        int cc
        void* cubin
        size_t cubin_size
        CUlibrary library
        CUkernel kernel

    cdef CUresult cccl_device_binary_search_build(
        cccl_device_binary_search_build_result_t*,
        cccl_binary_search_mode_t,
        cccl_iterator_t,
        cccl_iterator_t,
        cccl_iterator_t,
        cccl_op_t,
        int, int, const char*, const char*, const char*, const char*
    ) nogil

    cdef CUresult cccl_device_binary_search(
        cccl_device_binary_search_build_result_t,
        cccl_iterator_t,
        uint64_t,
        cccl_iterator_t,
        uint64_t,
        cccl_iterator_t,
        cccl_op_t,
        CUstream
    ) nogil

    cdef CUresult cccl_device_binary_search_cleanup(
        cccl_device_binary_search_build_result_t *build_ptr
    ) nogil


cdef class DeviceBinarySearchBuildResult:
    cdef cccl_device_binary_search_build_result_t build_data

    def __dealloc__(DeviceBinarySearchBuildResult self):
        cdef CUresult status = -1
        with nogil:
            status = cccl_device_binary_search_cleanup(&self.build_data)
        if (status != 0):
            print(f"Return code {status} encountered during binary_search result cleanup")

    def __cinit__(
        DeviceBinarySearchBuildResult self,
        cccl_binary_search_mode_t mode,
        Iterator d_data,
        Iterator d_values,
        Iterator d_out,
        Op op,
        CommonData common_data
    ):
        cdef CUresult status = -1
        cdef int cc_major = common_data.get_cc_major()
        cdef int cc_minor = common_data.get_cc_minor()
        cdef const char *cub_path = common_data.cub_path_get_c_str()
        cdef const char *thrust_path = common_data.thrust_path_get_c_str()
        cdef const char *libcudacxx_path = common_data.libcudacxx_path_get_c_str()
        cdef const char *ctk_path = common_data.ctk_path_get_c_str()

        memset(&self.build_data, 0, sizeof(cccl_device_binary_search_build_result_t))
        with nogil:
            status = cccl_device_binary_search_build(
                &self.build_data,
                mode,
                d_data.iter_data,
                d_values.iter_data,
                d_out.iter_data,
                op.op_data,
                cc_major,
                cc_minor,
                cub_path,
                thrust_path,
                libcudacxx_path,
                ctk_path,
            )
        if status != 0:
            raise RuntimeError(
                f"Failed building binary_search, error code: {status}"
            )

    cpdef void compute(
        DeviceBinarySearchBuildResult self,
        Iterator d_data,
        size_t num_items,
        Iterator d_values,
        size_t num_values,
        Iterator d_out,
        Op op,
        stream
    ):
        cdef CUresult status = -1
        cdef CUstream c_stream = <CUstream><uintptr_t>(stream) if stream else NULL

        with nogil:
            status = cccl_device_binary_search(
                self.build_data,
                d_data.iter_data,
                <uint64_t>num_items,
                d_values.iter_data,
                <uint64_t>num_values,
                d_out.iter_data,
                op.op_data,
                c_stream
            )
        if status != 0:
            raise RuntimeError(
                f"Failed executing binary_search, error code: {status}"
            )

    def _get_cubin(self):
        return PyBytes_FromStringAndSize(
            <const char*>self.build_data.cubin,
            self.build_data.cubin_size
        )


# ----------------------------------
# DeviceThreeWayPartitionBuildResult
# ----------------------------------
cdef extern from "cccl/c/three_way_partition.h":
    cdef struct cccl_device_three_way_partition_build_result_t 'cccl_device_three_way_partition_build_result_t':
        const char* cubin
        size_t cubin_size

    cdef CUresult cccl_device_three_way_partition_build(
        cccl_device_three_way_partition_build_result_t *build_ptr,
        cccl_iterator_t d_in,
        cccl_iterator_t d_first_part_out,
        cccl_iterator_t d_second_part_out,
        cccl_iterator_t d_unselected_out,
        cccl_iterator_t d_num_selected_out,
        cccl_op_t select_first_part_op,
        cccl_op_t select_second_part_op,
        int, int, const char *, const char *, const char *, const char *
    ) nogil

    CUresult cccl_device_three_way_partition(
        cccl_device_three_way_partition_build_result_t build,
        void* d_temp_storage,
        size_t* temp_storage_bytes,
        cccl_iterator_t d_in,
        cccl_iterator_t d_first_part_out,
        cccl_iterator_t d_second_part_out,
        cccl_iterator_t d_unselected_out,
        cccl_iterator_t d_num_selected_out,
        cccl_op_t select_first_part_op,
        cccl_op_t select_second_part_op,
        int64_t num_items,
        CUstream stream
    ) nogil

    cdef CUresult cccl_device_three_way_partition_cleanup(
        cccl_device_three_way_partition_build_result_t *build_ptr
    ) nogil


cdef class DeviceThreeWayPartitionBuildResult:
    cdef cccl_device_three_way_partition_build_result_t build_data

    def __dealloc__(DeviceThreeWayPartitionBuildResult self):
        cdef CUresult status = -1
        with nogil:
            status = cccl_device_three_way_partition_cleanup(&self.build_data)
        if (status != 0):
            print(f"Return code {status} encountered during three_way_partition result cleanup")


    def __cinit__(
        DeviceThreeWayPartitionBuildResult self,
        Iterator d_in,
        Iterator d_first_part_out,
        Iterator d_second_part_out,
        Iterator d_unselected_out,
        Iterator d_num_selected_out,
        Op select_first_part_op,
        Op select_second_part_op,
        CommonData common_data
    ):
        cdef CUresult status = -1
        cdef int cc_major = common_data.get_cc_major()
        cdef int cc_minor = common_data.get_cc_minor()
        cdef const char *cub_path = common_data.cub_path_get_c_str()
        cdef const char *thrust_path = common_data.thrust_path_get_c_str()
        cdef const char *libcudacxx_path = common_data.libcudacxx_path_get_c_str()
        cdef const char *ctk_path = common_data.ctk_path_get_c_str()

        memset(&self.build_data, 0, sizeof(cccl_device_three_way_partition_build_result_t))
        with nogil:
            status = cccl_device_three_way_partition_build(
                &self.build_data,
                d_in.iter_data,
                d_first_part_out.iter_data,
                d_second_part_out.iter_data,
                d_unselected_out.iter_data,
                d_num_selected_out.iter_data,
                select_first_part_op.op_data,
                select_second_part_op.op_data,
                cc_major,
                cc_minor,
                cub_path,
                thrust_path,
                libcudacxx_path,
                ctk_path,
            )
        if status != 0:
            raise RuntimeError(
                f"Failed building three_way_partition, error code: {status}"
            )

    cpdef size_t compute(
        DeviceThreeWayPartitionBuildResult self,
        temp_storage_ptr,
        temp_storage_bytes,
        Iterator d_in,
        Iterator d_first_part_out,
        Iterator d_second_part_out,
        Iterator d_unselected_out,
        Iterator d_num_selected_out,
        Op select_first_part_op,
        Op select_second_part_op,
        size_t num_items,
        stream
    ):
        cdef CUresult status = -1
        cdef void *storage_ptr = (<void *><uintptr_t>temp_storage_ptr) if temp_storage_ptr else NULL
        cdef size_t storage_sz = <size_t>temp_storage_bytes
        cdef CUstream c_stream = <CUstream><uintptr_t>(stream) if stream else NULL

        with nogil:
            status = cccl_device_three_way_partition(
                self.build_data,
                storage_ptr,
                &storage_sz,
                d_in.iter_data,
                d_first_part_out.iter_data,
                d_second_part_out.iter_data,
                d_unselected_out.iter_data,
                d_num_selected_out.iter_data,
                select_first_part_op.op_data,
                select_second_part_op.op_data,
                <uint64_t>num_items,
                c_stream
            )
        if status != 0:
            raise RuntimeError(
                f"Failed executing three_way_partition, error code: {status}"
            )
        return storage_sz

    def _get_cubin(self):
        return PyBytes_FromStringAndSize(
            <const char*>self.build_data.cubin,
            self.build_data.cubin_size
        )


# -------------------
# DeviceSegmentedSort
# -------------------

cdef extern from "cccl/c/segmented_sort.h":
    cdef struct cccl_device_segmented_sort_build_result_t 'cccl_device_segmented_sort_build_result_t':
        const char* cubin
        size_t cubin_size

    cdef CUresult cccl_device_segmented_sort_build(
        cccl_device_segmented_sort_build_result_t *build_ptr,
        cccl_sort_order_t sort_order,
        cccl_iterator_t d_keys_in,
        cccl_iterator_t d_keys_out,
        cccl_iterator_t begin_offset_in,
        cccl_iterator_t end_offset_in,
        int, int, const char *, const char *, const char *, const char *
    ) nogil

    cdef CUresult cccl_device_segmented_sort(
        cccl_device_segmented_sort_build_result_t build,
        void* d_temp_storage,
        size_t* temp_storage_bytes,
        cccl_iterator_t d_keys_in,
        cccl_iterator_t d_keys_out,
        cccl_iterator_t d_values_in,
        cccl_iterator_t d_values_out,
        int64_t num_items,
        int64_t num_segments,
        cccl_iterator_t start_offset_in,
        cccl_iterator_t end_offset_in,
        bint is_overwrite_okay,
        int* selector,
        CUstream stream
    ) nogil

    cdef CUresult cccl_device_segmented_sort_cleanup(
        cccl_device_segmented_sort_build_result_t* build_ptr
    ) nogil

cdef class DeviceSegmentedSortBuildResult:
    cdef cccl_device_segmented_sort_build_result_t build_data

    def __dealloc__(DeviceSegmentedSortBuildResult self):
        cdef CUresult status = -1
        with nogil:
            status = cccl_device_segmented_sort_cleanup(&self.build_data)
        if (status != 0):
            print(f"Return code {status} encountered during segmented_sort result cleanup")

    def __cinit__(
        DeviceSegmentedSortBuildResult self,
        cccl_sort_order_t order,
        Iterator d_keys_in,
        Iterator d_values_in,
        Iterator begin_offset_in,
        Iterator end_offset_in,
        CommonData common_data,
    ):
        cdef CUresult status = -1
        cdef int cc_major = common_data.get_cc_major()
        cdef int cc_minor = common_data.get_cc_minor()
        cdef const char *cub_path = common_data.cub_path_get_c_str()
        cdef const char *thrust_path = common_data.thrust_path_get_c_str()
        cdef const char *libcudacxx_path = common_data.libcudacxx_path_get_c_str()
        cdef const char *ctk_path = common_data.ctk_path_get_c_str()

        memset(&self.build_data, 0, sizeof(cccl_device_segmented_sort_build_result_t))
        with nogil:
            status = cccl_device_segmented_sort_build(
                &self.build_data,
                order,
                d_keys_in.iter_data,
                d_values_in.iter_data,
                begin_offset_in.iter_data,
                end_offset_in.iter_data,
                cc_major,
                cc_minor,
                cub_path,
                thrust_path,
                libcudacxx_path,
                ctk_path,
            )
        if status != 0:
            raise RuntimeError(
                f"Failed building segmented_sort, error code: {status}"
            )

    cpdef tuple compute(
        DeviceSegmentedSortBuildResult self,
        temp_storage_ptr,
        temp_storage_bytes,
        Iterator d_keys_in,
        Iterator d_keys_out,
        Iterator d_values_in,
        Iterator d_values_out,
        size_t num_items,
        size_t num_segments,
        Iterator start_offset_in,
        Iterator end_offset_in,
        bint is_overwrite_okay,
        selector,
        stream
    ):
        cdef CUresult status = -1
        cdef void *storage_ptr = (<void *><size_t>temp_storage_ptr) if temp_storage_ptr else NULL
        cdef size_t storage_sz = <size_t>temp_storage_bytes
        cdef int selector_int = <int>selector
        cdef CUstream c_stream = <CUstream><size_t>(stream) if stream else NULL

        with nogil:
            status = cccl_device_segmented_sort(
                self.build_data,
                storage_ptr,
                &storage_sz,
                d_keys_in.iter_data,
                d_keys_out.iter_data,
                d_values_in.iter_data,
                d_values_out.iter_data,
                <uint64_t>num_items,
                <uint64_t>num_segments,
                start_offset_in.iter_data,
                end_offset_in.iter_data,
                is_overwrite_okay,
                &selector_int,
                c_stream
            )

        if status != 0:
            raise RuntimeError(
                f"Failed executing segmented_sort, error code: {status}"
            )
        return <object>storage_sz, <object>selector_int


    def _get_cubin(self):
        return PyBytes_FromStringAndSize(
            <const char*>self.build_data.cubin,
            self.build_data.cubin_size
        )



# Copyright (c) 2024, NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
#
#
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

from __future__ import annotations

import functools
import types
from typing import Any, Callable, Hashable

import numpy as np

try:
    from cuda.core import Device
except ImportError:
    from cuda.core.experimental import Device

from ._utils.protocols import get_dtype, get_shape, is_device_array
from .struct import _Struct

# Registry thet maps type -> key function for extracting cache key
# from a value of that type.
_KEY_FUNCTIONS: dict[type, Callable[[Any], Hashable]] = {}


def _type_fqn(v):
    # fully-qualified type name to distinguish np.ndarray from cp.ndarray from GpuStruct
    return f"{type(v).__module__}.{type(v).__name__}"


def _key_for(value: Any) -> Hashable:
    """
    Extract a cache key from a value using the registered KEY_FUNCTIONS.

    This function checks the type of the value and delegates to the
    appropriate registered keyer. Falls back to using the value
    directly if no keyer is registered.

    Args:
        value: The value to extract a cache key from

    Returns:
        A hashable cache key
    """
    # Handle sequences (lists, tuples) by recursively converting to tuple
    if isinstance(value, (list, tuple)):
        return tuple(_key_for(item) for item in value)

    # Check for exact type match first
    value_type = type(value)
    if value_type in _KEY_FUNCTIONS:
        return _KEY_FUNCTIONS[value_type](value)

    # DeviceArrayLike is not a runtime-checkable protocol, so
    # we cannot isinstance() with it.
    if is_device_array(value):
        return (_type_fqn(value), get_dtype(value))

    # Check for instance match (handles inheritance)
    for registered_type, keyer in _KEY_FUNCTIONS.items():
        if isinstance(value, registered_type):
            return keyer(value)

    # Fallback: use value directly (assumes it's hashable)
    return value


def _make_cache_key_from_args(*args, **kwargs) -> tuple:
    """
    Create a cache key from function arguments.

    Args:
        *args: Positional arguments
        **kwargs: Keyword arguments

    Returns:
        A tuple containing the extracted cache keys
    """

    positional_keys = tuple(_key_for(arg) for arg in args)

    # Sort kwargs by key name for consistent ordering
    if kwargs:
        sorted_kwargs = sorted(kwargs.items())
        kwarg_keys = tuple((k, _key_for(v)) for k, v in sorted_kwargs)
        return positional_keys + (kwarg_keys,)

    return positional_keys


# Central registry of all algorithm caches
_cache_registry: dict[str, object] = {}


class _CacheWithRegisteredKeyFunctions:
    """
    Decorator to cache the result of the decorated function.

    The cache key is automatically computed from the decorated function's
    arguments using the registered key functions.
    """

    def __call__(self, func: Callable) -> Callable:
        """
        Decorator to cache the result of the decorated function.

        Args:
            func: The function whose result is to be cached.

        Notes
        -----
        The CUDA compute capability of the current device is appended to
        the cache key.
        """
        cache: dict = {}

        @functools.wraps(func)
        def inner(*args, **kwargs):
            cc = Device().compute_capability
            user_cache_key = _make_cache_key_from_args(*args, **kwargs)
            cache_key = (user_cache_key, tuple(cc))
            if cache_key not in cache:
                result = func(*args, **kwargs)
                cache[cache_key] = result
            return cache[cache_key]

        inner.cache_clear = cache.clear  # type: ignore[attr-defined]

        # Register the cache in the central registry
        _cache_registry[func.__qualname__] = inner

        return inner

    def register(self, type_: type, key_function: Callable[[Any], Hashable]) -> None:
        """
        Register a key function for a specific type.

        A key function extracts a hashable cache key from a value.

        Args:
            type_: The type to register
            key_function: A callable that takes an instance of type_ and
                returns a hashable cache key
        """
        _KEY_FUNCTIONS[type_] = key_function


cache_with_registered_key_functions = _CacheWithRegisteredKeyFunctions()


def _make_hashable(value):
    # duck-type check for numba.cuda.CUDADispatcher:
    if hasattr(value, "py_func") and callable(value.py_func):
        return CachableFunction(value.py_func)
    elif is_device_array(value):
        # Ops with device arrays in globals/closures will be handled
        # by stateful op machinery, which enables updating the state
        # (pointers). Thus, we only cache on the dtype and shape of
        # the referenced array, but not its pointer.
        return (get_dtype(value), get_shape(value))
    elif isinstance(value, (list, tuple)):
        return tuple(_make_hashable(v) for v in value)
    elif isinstance(value, dict):
        return tuple(
            sorted((_make_hashable(k), _make_hashable(v)) for k, v in value.items())
        )
    else:
        return id(value)


def clear_all_caches():
    """
    Clear all algorithm caches.

    This function clears all cached algorithm build results, forcing
    recompilation on the next invocation. Useful for benchmarking
    compilation time.

    Example
    -------
    >>> import cuda.compute
    >>> cuda.compute.clear_all_caches()
    """
    for cached_func in _cache_registry.values():
        cached_func.cache_clear()


class CachableFunction:
    """
    A type that wraps a function and provides custom comparison
    (__eq__) and hash (__hash__) implementations.

    The purpose of this class is to enable caching and comparison of
    functions based on their bytecode, constants, and closures, while
    ignoring other attributes such as their names or docstrings.
    """

    # TODO: eventually, move this class to _jit.py as it only
    # has to do with caching of Python callables that will be
    # JIT compiled.
    def __init__(self, func):
        self._func = func

        closure = func.__closure__ if func.__closure__ is not None else []
        contents = []
        # Make closure contents hashable
        for cell in closure:
            contents.append(_make_hashable(cell.cell_contents))
        self._identity = (
            func.__name__,
            func.__code__.co_code,
            func.__code__.co_consts,
            tuple(contents),
            tuple(
                # if `name` is found in __globals__, try and hash
                # the referenced object. If `name` is not found in
                # __globals__, (e.g., `name` is part of a dotted
                # name like `np.argmax`), for caching purposes we
                # use the hash of the name itself. Assumes numba
                # known how to interpret the dotted name at JIT
                # time.
                _make_hashable(func.__globals__.get(name, name))
                for name in func.__code__.co_names
            ),
        )

    def __eq__(self, other):
        return self._identity == other._identity

    def __hash__(self):
        return hash(self._identity)

    def __repr__(self):
        return str(self._func)


# Register keyers for built-in types

cache_with_registered_key_functions.register(
    np.ndarray, lambda arr: ("numpy.ndarray", arr.dtype)
)
cache_with_registered_key_functions.register(
    types.FunctionType, lambda fn: CachableFunction(fn)
)
cache_with_registered_key_functions.register(_Struct, lambda v: (_type_fqn(v), v.dtype))



from __future__ import annotations

import re

__all__ = ["sanitize_identifier"]


def sanitize_identifier(name: str) -> str:
    """Sanitize a name to be a valid Python/LLVM identifier.

    This replaces any character that isn't alphanumeric or underscore with
    an underscore. This is needed because:
    - Lambda functions have __name__ = "<lambda>" which contains angle brackets
    - Python identifiers and LLVM/NVVM global names don't allow special characters

    Args:
        name: The name to sanitize (e.g., function __name__)

    Returns:
        A sanitized name safe for use as a Python identifier or LLVM symbol
    """
    return re.sub(r"[^a-zA-Z0-9_]", "_", name)



# Copyright (c) 2026, NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
#
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

"""
C++ code generation and compilation infrastructure.
"""

from __future__ import annotations

import functools

from cuda.cccl import get_include_paths
from cuda.core import Device, Program, ProgramOptions

from ._bindings import TypeEnum


def _get_arch_string() -> str:
    """Get the compute capability string for the current device."""

    device = Device()
    cc_major, cc_minor = device.compute_capability
    return f"sm_{cc_major}{cc_minor}"


@functools.lru_cache(maxsize=1)
def _get_include_paths() -> list[str]:
    """Get include paths for CCCL headers."""
    paths = get_include_paths().as_tuple()
    return [p for p in paths if p is not None]


@functools.lru_cache(maxsize=256)
def compile_cpp_to_ltoir(
    source: str,
    arch: str | None = None,
) -> bytes:
    """
    Compile C++ source code to LTOIR.

    Args:
        source: C++ source code string
        arch: Target architecture (e.g., "sm_80"). If None, uses current device.

    Returns:
        LTOIR bytes

    Example:
        source = '''
        extern "C" __device__ void my_add(void* a, void* b, void* result) {
            *static_cast<int*>(result) = *static_cast<int*>(a) + *static_cast<int*>(b);
        }
        '''
        ltoir = compile_cpp_to_ltoir(source)
    """

    if arch is None:
        arch = _get_arch_string()

    # Get include paths
    include_paths = _get_include_paths()

    # Configure compilation options for LTO
    opts = ProgramOptions(
        arch=arch,
        relocatable_device_code=True,
        link_time_optimization=True,
        std="c++20",
        include_path=include_paths,
    )

    # Compile to LTOIR
    program = Program(source, "c++", options=opts)
    result = program.compile("ltoir")

    return result.code


def cpp_type_from_descriptor(type_desc) -> str | None:
    """
    Get the C++ type name from a TypeDescriptor.

    Important: for efficiency, this function returns None
    for non-primitive types. Callers must take care
    to handle that case appropriately.
    """
    # Map TypeEnum to C++ types
    type_map = {
        TypeEnum.INT8: "int8_t",
        TypeEnum.INT16: "int16_t",
        TypeEnum.INT32: "int32_t",
        TypeEnum.INT64: "int64_t",
        TypeEnum.UINT8: "uint8_t",
        TypeEnum.UINT16: "uint16_t",
        TypeEnum.UINT32: "uint32_t",
        TypeEnum.UINT64: "uint64_t",
        TypeEnum.FLOAT16: "__half",
        TypeEnum.FLOAT32: "float",
        TypeEnum.FLOAT64: "double",
        TypeEnum.BOOLEAN: "bool",
        TypeEnum.STORAGE: None,
    }
    return type_map[type_desc.info.typenum]


def make_variable_declaration(type_desc, name: str) -> str:
    """
    Generate a C++ variable declaration, like "int32_t temp;"
    or "alignas(8) char temp[16];"
    """
    cpp_type = cpp_type_from_descriptor(type_desc)
    if cpp_type is not None:
        return f"{cpp_type} {name};"
    # STORAGE type - use aligned char array
    return f"alignas({type_desc.alignment}) char {name}[{type_desc.size}];"



# Copyright (c) 2024, NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
#
#
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

from __future__ import annotations

from ._binary_search import lower_bound as lower_bound
from ._binary_search import make_lower_bound as make_lower_bound
from ._binary_search import make_upper_bound as make_upper_bound
from ._binary_search import upper_bound as upper_bound
from ._histogram import histogram_even as histogram_even
from ._histogram import make_histogram_even as make_histogram_even
from ._reduce import make_reduce_into as make_reduce_into
from ._reduce import reduce_into as reduce_into
from ._scan import exclusive_scan as exclusive_scan
from ._scan import inclusive_scan as inclusive_scan
from ._scan import make_exclusive_scan as make_exclusive_scan
from ._scan import make_inclusive_scan as make_inclusive_scan
from ._segmented_reduce import make_segmented_reduce as make_segmented_reduce
from ._segmented_reduce import segmented_reduce
from ._select import make_select as make_select
from ._select import select as select
from ._sort import DoubleBuffer, SortOrder
from ._sort import make_merge_sort as make_merge_sort
from ._sort import make_radix_sort as make_radix_sort
from ._sort import make_segmented_sort as make_segmented_sort
from ._sort import merge_sort as merge_sort
from ._sort import radix_sort as radix_sort
from ._sort import segmented_sort as segmented_sort
from ._three_way_partition import make_three_way_partition as make_three_way_partition
from ._three_way_partition import three_way_partition as three_way_partition
from ._transform import binary_transform, unary_transform
from ._transform import make_binary_transform as make_binary_transform
from ._transform import make_unary_transform as make_unary_transform
from ._unique_by_key import make_unique_by_key as make_unique_by_key
from ._unique_by_key import unique_by_key as unique_by_key

__all__ = [
    "reduce_into",
    "make_reduce_into",
    "lower_bound",
    "make_lower_bound",
    "upper_bound",
    "make_upper_bound",
    "inclusive_scan",
    "make_inclusive_scan",
    "exclusive_scan",
    "make_exclusive_scan",
    "unary_transform",
    "make_unary_transform",
    "binary_transform",
    "make_binary_transform",
    "histogram_even",
    "make_histogram_even",
    "merge_sort",
    "make_merge_sort",
    "radix_sort",
    "make_radix_sort",
    "segmented_reduce",
    "make_segmented_reduce",
    "unique_by_key",
    "make_unique_by_key",
    "segmented_sort",
    "make_segmented_sort",
    "three_way_partition",
    "make_three_way_partition",
    "select",
    "make_select",
    "DoubleBuffer",
    "SortOrder",
]



# Copyright (c) 2026, NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
#
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

"""DiscardIterator implementation."""

from __future__ import annotations

from .._bindings import Op, OpKind
from .._cpp_compile import compile_cpp_to_ltoir
from .._utils.protocols import get_dtype
from .._utils.temp_storage_buffer import TempStorageBuffer
from ..types import TypeDescriptor, from_numpy_dtype
from ._base import IteratorBase
from ._common import CUDA_PREAMBLE


class DiscardIterator(IteratorBase):
    """
    Iterator that discards all reads and writes.
    """

    def __init__(self, reference_iterator=None):
        """
        Create a discard iterator.

        Args:
            reference_iterator: Optional iterator or device array used to infer
                value_type/state_type. Defaults to a temporary byte buffer.
        """
        if reference_iterator is None:
            reference_iterator = TempStorageBuffer(1)

        self._reference_iterator = reference_iterator

        if hasattr(reference_iterator, "__cuda_array_interface__"):
            value_type = from_numpy_dtype(get_dtype(reference_iterator))
            state_bytes = bytes(value_type.dtype.itemsize)
        elif isinstance(reference_iterator, IteratorBase):
            value_type = reference_iterator.value_type
            if isinstance(value_type, TypeDescriptor):
                state_bytes = bytes(value_type.dtype.itemsize)
            else:
                state_bytes = bytes(value_type.info.size)
        else:
            raise TypeError("reference_iterator must be a device array or iterator")

        super().__init__(
            state_bytes=state_bytes,
            state_alignment=value_type.alignment,
            value_type=value_type,
        )

    def _make_advance_op(self) -> Op:
        symbol = self._make_advance_symbol()

        source = f"""{CUDA_PREAMBLE}

extern "C" __device__ void {symbol}(void*, void*) {{
}}
"""
        ltoir = compile_cpp_to_ltoir(source)
        return Op(
            operator_type=OpKind.STATELESS,
            name=symbol,
            ltoir=ltoir,
            extra_ltoirs=[],
        )

    def _make_input_deref_op(self) -> Op | None:
        symbol = self._make_input_deref_symbol()

        source = f"""{CUDA_PREAMBLE}

extern "C" __device__ void {symbol}(void*, void*) {{
}}
"""
        ltoir = compile_cpp_to_ltoir(source)
        return Op(
            operator_type=OpKind.STATELESS,
            name=symbol,
            ltoir=ltoir,
            extra_ltoirs=[],
        )

    def _make_output_deref_op(self) -> Op | None:
        symbol = self._make_output_deref_symbol()

        source = f"""{CUDA_PREAMBLE}

extern "C" __device__ void {symbol}(void*, void*) {{
}}
"""
        ltoir = compile_cpp_to_ltoir(source)
        return Op(
            operator_type=OpKind.STATELESS,
            name=symbol,
            ltoir=ltoir,
            extra_ltoirs=[],
        )

    def __add__(self, offset: int) -> "DiscardIterator":
        return DiscardIterator(self._reference_iterator)



# Copyright (c) 2025, NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
#
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

"""
CUDA version detection utilities shared across the cccl package.
"""

from typing import Optional

import cuda.bindings


def detect_cuda_version() -> Optional[int]:
    cuda_version = cuda.bindings.__version__
    return int(cuda_version.split(".")[0])


def get_recommended_extra(cuda_version: Optional[int]) -> str:
    """Get the recommended pip extra for the detected CUDA version."""
    if cuda_version == 13:
        return "cu13"
    else:
        return "cu12"



# Copyright (c) 2024, NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
#
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception


import numba

from .._common import make_binary_tempfile
from .._types import (
    Algorithm,
    Dependency,
    DependentReference,
    Invocable,
    Pointer,
    TemplateParameter,
)


def make_exclusive_sum(dtype, threads_in_warp=32):
    """
    Creates an exclusive warp-wide prefix sum primitive using addition (+)
    as the scan operator. The value of 0 is applied as the initial value
    and is assigned to the output in lane :sub:`0`.

    Example:
        The code snippet below illustrates an exclusive prefix sum of
        32 integer items:

        .. literalinclude:: ../../python/cuda_cccl/tests/coop/test_warp_scan_api.py
            :language: python
            :dedent:
            :start-after: example-begin imports
            :end-before: example-end imports

        The following snippet shows how to invoke the returned
        ``warp_exclusive_sum`` primitive:

        .. literalinclude:: ../../python/cuda_cccl/tests/coop/test_warp_scan_api.py
            :language: python
            :dedent:
            :start-after: example-begin exclusive-sum
            :end-before: example-end exclusive-sum

        Suppose the set of input ``thread_data`` across the warp of
        threads is ``{ [1, 1, 1, 1], [1, 1, 1, 1], ..., [1, 1, 1, 1] }``.
        The corresponding output ``thread_data`` in those threads will be
        ``{ [0, 1, 2, 3], [4, 5, 6, 7], ..., [28, 29, 30, 31] }``.

    Args:
        dtype: Data type being scanned
        threads_in_warp: The number of threads in a warp

    Returns:
        A callable object that can be linked to and invoked from a CUDA
        kernel
    """
    template = Algorithm(
        "WarpScan",
        "ExclusiveSum",
        "warp_scan",
        ["cub/warp/warp_scan.cuh"],
        [TemplateParameter("T"), TemplateParameter("VIRTUAL_WARP_THREADS")],
        [
            [
                Pointer(numba.uint8),
                DependentReference(Dependency("T")),
                DependentReference(Dependency("T"), True),
            ]
        ],
        fake_return=True,
    )
    specialization = template.specialize(
        {"T": dtype, "VIRTUAL_WARP_THREADS": threads_in_warp}
    )
    return Invocable(
        temp_files=[
            make_binary_tempfile(ltoir, ".ltoir")
            for ltoir in specialization.get_lto_ir(threads=threads_in_warp)
        ],
        temp_storage_bytes=specialization.temp_storage_bytes,
        temp_storage_alignment=specialization.temp_storage_alignment,
        algorithm=specialization,
    )



# Copyright (c) 2024, NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
#
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

from ._warp_merge_sort import make_merge_sort_keys
from ._warp_reduce import make_reduce, make_sum
from ._warp_scan import make_exclusive_sum

__all__ = ["make_exclusive_sum", "make_reduce", "make_sum", "make_merge_sort_keys"]



from __future__ import annotations

from .._utils.protocols import is_device_array
from ._base import IteratorBase

CUDA_PREAMBLE = """#include <cuda/std/cstdint>
#include <cuda_fp16.h>
#include <cuda/std/cstring>
using namespace cuda::std;
"""


def ensure_iterator(obj):
    """Wrap array in PointerIterator if needed."""
    from ._pointer import PointerIterator

    if isinstance(obj, IteratorBase):
        return obj
    if is_device_array(obj):
        return PointerIterator(obj)
    raise TypeError("Expected an iterator or a device array")



# Copyright (c) 2026, NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
#
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
from __future__ import annotations

import inspect
from typing import get_type_hints

import numpy as np

from ._bindings import TypeEnum, TypeInfo

_ENUM_TO_DTYPE: dict[TypeEnum, np.dtype] = {
    TypeEnum.INT8: np.dtype("int8"),
    TypeEnum.INT16: np.dtype("int16"),
    TypeEnum.INT32: np.dtype("int32"),
    TypeEnum.INT64: np.dtype("int64"),
    TypeEnum.UINT8: np.dtype("uint8"),
    TypeEnum.UINT16: np.dtype("uint16"),
    TypeEnum.UINT32: np.dtype("uint32"),
    TypeEnum.UINT64: np.dtype("uint64"),
    TypeEnum.FLOAT16: np.dtype("float16"),
    TypeEnum.FLOAT32: np.dtype("float32"),
    TypeEnum.FLOAT64: np.dtype("float64"),
    TypeEnum.BOOLEAN: np.dtype("bool"),
}


class TypeDescriptor:
    def __init__(self, size: int, alignment: int, type_enum: TypeEnum):
        self._type_info = TypeInfo(size, alignment, type_enum)
        self._dtype = _ENUM_TO_DTYPE.get(type_enum, np.dtype(f"V{size}"))

    @property
    def info(self) -> TypeInfo:
        """Return the TypeInfo for this type."""
        return self._type_info

    @property
    def size(self):
        return self._type_info.size

    @property
    def alignment(self):
        return self._type_info.alignment

    @property
    def dtype(self) -> np.dtype:
        """Return the numpy dtype for this type."""
        return self._dtype

    def pointer(self) -> "PointerTypeDescriptor":
        """Create a pointer type to this type."""
        return PointerTypeDescriptor(self)

    def __repr__(self) -> str:
        return f"TypeDescriptor({self._dtype})"

    def __eq__(self, other):
        if not isinstance(other, TypeDescriptor):
            return False
        return self._dtype == other._dtype

    def __hash__(self):
        return hash(self._dtype)


class StructTypeDescriptor(TypeDescriptor):
    def __init__(
        self,
        fields: dict[str, "TypeDescriptor"],
        name: str = "AnonStruct",
    ):
        dtype = _build_struct_dtype(fields)
        if not dtype.isalignedstruct:
            raise ValueError(f"dtype {dtype} must be aligned")
        # Structs use STORAGE type enum
        super().__init__(dtype.itemsize, dtype.alignment, TypeEnum.STORAGE)
        self._dtype = dtype
        self._fields = fields
        self._name = name

    def __repr__(self) -> str:
        return f"StructTypeDescriptor({self._name}, dtype={self._dtype})"

    @property
    def name(self) -> str:
        return self._name

    @property
    def fields(self) -> dict[str, "TypeDescriptor"]:
        return self._fields

    def layout_key(self) -> tuple[tuple[str, "TypeDescriptor"], ...]:
        """Return a stable, hashable key for this struct layout."""
        return tuple(self._fields.items())

    def __eq__(self, other):
        if not isinstance(other, StructTypeDescriptor):
            return False
        # Compare by fields (TypeDescriptors) to correctly distinguish structs
        # with different pointer pointee types, which have identical numpy dtypes
        # (both uint64) but different semantics.
        # Must compare as ordered items because field order matters for struct layout.
        return list(self._fields.items()) == list(other._fields.items())

    def __hash__(self):
        return hash(tuple(self._fields.items()))


class PointerTypeDescriptor(TypeDescriptor):
    def __init__(self, pointee: TypeDescriptor):
        # Pointer is 8 bytes, 8-byte aligned, represented as uint64
        super().__init__(8, 8, TypeEnum.UINT64)
        self._pointee = pointee

    @property
    def pointee(self) -> TypeDescriptor:
        """Return the type this pointer points to."""
        return self._pointee

    def __repr__(self) -> str:
        return f"PointerTypeDescriptor({self._pointee})"

    def __eq__(self, other):
        if not isinstance(other, PointerTypeDescriptor):
            return False
        return self._pointee == other._pointee

    def __hash__(self):
        return hash(("PointerTypeDescriptor", self._pointee))


def _build_struct_dtype(
    fields: dict[str, TypeDescriptor],
) -> np.dtype:
    dtype_list = []
    for field_name, field_type in fields.items():
        dtype_list.append((field_name, field_type.dtype))
    return np.dtype(dtype_list, align=True)


def struct(
    fields: dict[str, TypeDescriptor],
    name: str = "AnonStruct",
) -> StructTypeDescriptor:
    """Create a type descriptor for a struct."""
    return StructTypeDescriptor(fields, name=name)


def pointer(pointee: TypeDescriptor) -> PointerTypeDescriptor:
    """
    Create a pointer to the given type.
    """
    return PointerTypeDescriptor(pointee)


def from_numpy_dtype(dtype: np.dtype | type) -> TypeDescriptor:
    """
    Convert a numpy dtype (or numpy type) to a TypeDescriptor.

    Handles POD types and structured dtypes (recursively for nested structs).
    """
    dtype = np.dtype(dtype)

    # Check if it's a known POD type
    td = _DTYPE_TO_TD.get(dtype)
    if td is not None:
        return td

    # Handle structured dtypes (structs)
    if dtype.names is not None:
        fields: dict[str, TypeDescriptor] = {}
        assert dtype.fields is not None
        for name in dtype.names:
            field_info = dtype.fields[name]
            field_dtype = field_info[0]
            fields[name] = from_numpy_dtype(field_dtype)
        return struct(fields)  # type: ignore[arg-type]

    # Some other NumPy type (e.g., complex64) for which we don't
    # have a specific TypeDescriptor. Use STORAGE and preserve the original dtype.
    td = TypeDescriptor(dtype.itemsize, dtype.alignment, TypeEnum.STORAGE)
    td._dtype = dtype
    return td


# Signed integer types
int8 = TypeDescriptor(1, 1, TypeEnum.INT8)
int16 = TypeDescriptor(2, 2, TypeEnum.INT16)
int32 = TypeDescriptor(4, 4, TypeEnum.INT32)
int64 = TypeDescriptor(8, 8, TypeEnum.INT64)

# Unsigned integer types
uint8 = TypeDescriptor(1, 1, TypeEnum.UINT8)
uint16 = TypeDescriptor(2, 2, TypeEnum.UINT16)
uint32 = TypeDescriptor(4, 4, TypeEnum.UINT32)
uint64 = TypeDescriptor(8, 8, TypeEnum.UINT64)

# Floating point types
float16 = TypeDescriptor(2, 2, TypeEnum.FLOAT16)
float32 = TypeDescriptor(4, 4, TypeEnum.FLOAT32)
float64 = TypeDescriptor(8, 8, TypeEnum.FLOAT64)

# Boolean
boolean = TypeDescriptor(1, 1, TypeEnum.BOOLEAN)


# Mapping from numpy dtype to TypeDescriptor for POD types
_DTYPE_TO_TD: dict[np.dtype, TypeDescriptor] = {
    np.dtype("int8"): int8,
    np.dtype("int16"): int16,
    np.dtype("int32"): int32,
    np.dtype("int64"): int64,
    np.dtype("uint8"): uint8,
    np.dtype("uint16"): uint16,
    np.dtype("uint32"): uint32,
    np.dtype("uint64"): uint64,
    np.dtype("float16"): float16,
    np.dtype("float32"): float32,
    np.dtype("float64"): float64,
    np.dtype("bool"): boolean,
}


def to_ctypes_type(td: TypeDescriptor):
    """Convert a TypeDescriptor to a ctypes type."""
    return np.ctypeslib.as_ctypes_type(td.dtype)


def _annotation_to_type_descriptor(annotation):
    """
    Convert a type annotation to a TypeDescriptor.

    Handles:
    - TypeDescriptor: returns as-is
    - gpu_struct classes: returns their _type_descriptor
    - numpy dtypes/types: converts via from_numpy_dtype
    """
    from .struct import _is_struct_type

    if isinstance(annotation, TypeDescriptor):
        return annotation

    if _is_struct_type(annotation):
        return annotation._type_descriptor  # type: ignore[union-attr]

    # numpy dtype or type
    return from_numpy_dtype(np.dtype(annotation))


def signature_from_annotations(py_func):
    try:
        annotations = get_type_hints(py_func)
    except Exception:
        annotations = py_func.__annotations__
    spec = inspect.getfullargspec(py_func)
    arg_names = list(spec.args)

    input_tds = []
    # Try to get input types from annotations
    for name in arg_names:
        if name in annotations:
            input_tds.append(_annotation_to_type_descriptor(annotations[name]))
            break

    if "return" in annotations:
        output_td = _annotation_to_type_descriptor(annotations["return"])
    else:
        output_td = None

    return input_tds, output_td


__all__ = [
    "int8",
    "int16",
    "int32",
    "int64",
    "uint8",
    "uint16",
    "uint32",
    "uint64",
    "float16",
    "float32",
    "float64",
    "boolean",
    "struct",
    "pointer",
    "from_numpy_dtype",
    "to_ctypes_type",
]



# Copyright (c) 2024-2025, NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
#
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

from typing import TYPE_CHECKING, Callable, Literal, Tuple, Union

import numba

from .._common import (
    CUB_BLOCK_REDUCE_ALGOS,
    make_binary_tempfile,
    normalize_dim_param,
    normalize_dtype_param,
)
from .._types import (
    Algorithm,
    Dependency,
    DependentArray,
    DependentPythonOperator,
    DependentReference,
    Invocable,
    Pointer,
    TemplateParameter,
    Value,
    numba_type_to_wrapper,
)

if TYPE_CHECKING:
    import numpy as np


def _reduce(
    dtype: Union[str, type, "np.dtype", "numba.types.Type"],
    threads_per_block: Union[int, Tuple[int, int], Tuple[int, int, int]],
    items_per_thread: int,
    binary_op: Callable,
    algorithm: Literal["raking", "raking_commutative_only", "warp_reductions"],
    methods: dict = None,
) -> Callable:
    if algorithm not in CUB_BLOCK_REDUCE_ALGOS:
        raise ValueError(f"Unsupported algorithm: {algorithm}")

    if items_per_thread < 1:
        raise ValueError("items_per_thread must be greater than or equal to 1")

    dim = normalize_dim_param(threads_per_block)
    dtype = normalize_dtype_param(dtype)

    specialization_kwds = {
        "T": dtype,
        "BLOCK_DIM_X": dim[0],
        "ALGORITHM": CUB_BLOCK_REDUCE_ALGOS[algorithm],
        "BLOCK_DIM_Y": dim[1],
        "BLOCK_DIM_Z": dim[2],
    }

    template_parameters = [
        TemplateParameter("T"),
        TemplateParameter("BLOCK_DIM_X"),
        TemplateParameter("ALGORITHM"),
        TemplateParameter("BLOCK_DIM_Y"),
        TemplateParameter("BLOCK_DIM_Z"),
    ]

    if binary_op is None:
        cpp_method_name = "Sum"

        if items_per_thread == 1:
            parameters = [
                # Signatures:
                # T Sum(T);
                [
                    Pointer(numba.uint8),
                    DependentReference(Dependency("T")),
                    DependentReference(Dependency("T"), is_output=True),
                ],
                # T Sum(T, int num_valid);
                [
                    Pointer(numba.uint8),
                    DependentReference(Dependency("T")),
                    Value(numba.int32),
                    DependentReference(Dependency("T"), is_output=True),
                ],
            ]

        else:
            assert items_per_thread > 1, items_per_thread
            specialization_kwds["ITEMS_PER_THREAD"] = items_per_thread

            parameters = [
                # Signatures:
                # T Sum(T(&)[ITEMS_PER_THREAD]);
                [
                    Pointer(numba.uint8),
                    DependentArray(Dependency("T"), Dependency("ITEMS_PER_THREAD")),
                    DependentReference(Dependency("T"), is_output=True),
                ],
            ]

    else:
        cpp_method_name = "Reduce"
        specialization_kwds["Op"] = binary_op

        if items_per_thread == 1:
            parameters = [
                # Signatures:
                # T Reduce(T, Op);
                [
                    Pointer(numba.uint8),
                    DependentReference(Dependency("T")),
                    DependentPythonOperator(
                        Dependency("T"),
                        [Dependency("T"), Dependency("T")],
                        Dependency("Op"),
                    ),
                    DependentReference(Dependency("T"), is_output=True),
                ],
                # T Reduce(T, Op, int num_valid);
                [
                    Pointer(numba.uint8),
                    DependentReference(Dependency("T")),
                    DependentPythonOperator(
                        Dependency("T"),
                        [Dependency("T"), Dependency("T")],
                        Dependency("Op"),
                    ),
                    Value(numba.int32),
                    DependentReference(Dependency("T"), is_output=True),
                ],
            ]

        else:
            assert items_per_thread > 1, items_per_thread
            specialization_kwds["ITEMS_PER_THREAD"] = items_per_thread

            parameters = [
                # Signatures:
                # T Reduce(T(&)[ITEMS_PER_THREAD], Op);
                [
                    Pointer(numba.uint8),
                    DependentArray(Dependency("T"), Dependency("ITEMS_PER_THREAD")),
                    DependentPythonOperator(
                        Dependency("T"),
                        [Dependency("T"), Dependency("T")],
                        Dependency("Op"),
                    ),
                    DependentReference(Dependency("T"), is_output=True),
                ],
            ]

    template = Algorithm(
        "BlockReduce",
        cpp_method_name,
        "block_reduce",
        ["cub/block/block_reduce.cuh"],
        template_parameters,
        parameters,
        type_definitions=[numba_type_to_wrapper(dtype, methods=methods)],
    )

    specialization = template.specialize(specialization_kwds)
    return Invocable(
        temp_files=[
            make_binary_tempfile(ltoir, ".ltoir")
            for ltoir in specialization.get_lto_ir()
        ],
        temp_storage_bytes=specialization.temp_storage_bytes,
        temp_storage_alignment=specialization.temp_storage_alignment,
        algorithm=specialization,
    )


def make_reduce(
    dtype,
    threads_per_block,
    binary_op,
    items_per_thread=1,
    algorithm="warp_reductions",
    methods=None,
):
    """
    Creates a block-wide reduction primitive for thread :sub:`0` using
    the specified binary reduction functor.

    Returns a callable object that can be linked to and invoked from
    device code. It can be invoked with the following signatures:

    - `(item: dtype) -> dtype)`: Each thread contributes a single item to
      the reduction.
    - `(items: numba.types.Array) -> dtype`: Each thread contributes an
      array of items to the reduction. The array must contain at least
      `items_per_thread` items; only the first `items_per_thread` items
      are included in the reduction.
    - `(item: dtype, num_valid: int) -> dtype`: The first `num_valid`
      threads contribute a single item to the reduction. Items from all
      other threads are ignored.

    Args:
        dtype: Data type being reduced
        threads_per_block: Number of threads in a block. Can be an
            integer or a tuple of 2 or 3 integers.
        binary_op: Binary reduction function
        items_per_thread: The number of items each thread contributes to
            the reduction
        algorithm: Algorithm to use for the reduction (one of "raking",
            "raking_commutative_only", "warp_reductions")
        methods: A dict of methods for user-defined types

    Warning:
        The return value is undefined in threads other than thread :sub:`0`.

    Example:
        The code snippet below illustrates a max reduction of 128
        integer items that are partitioned across 128 threads.

        .. literalinclude:: ../../python/cuda_cccl/tests/coop/test_block_reduce_api.py
            :language: python
            :dedent:
            :start-after: example-begin imports
            :end-before: example-end imports

        The following snippet shows how to invoke the returned
        ``block_reduce`` primitive:

        .. literalinclude:: ../../python/cuda_cccl/tests/coop/test_block_reduce_api.py
            :language: python
            :dedent:
            :start-after: example-begin reduce
            :end-before: example-end reduce

        Suppose the set of inputs across the block of threads is
        ``{ 0, 1, 2, 3, ..., 127 }``. The output in thread :sub:`0`
        is ``{ 127 }``.
    """
    return _reduce(
        dtype=dtype,
        threads_per_block=threads_per_block,
        items_per_thread=items_per_thread,
        binary_op=binary_op,
        algorithm=algorithm,
        methods=methods,
    )


def make_sum(
    dtype,
    threads_per_block,
    items_per_thread=1,
    algorithm="warp_reductions",
    methods=None,
):
    """
    Creates a block-wide reduction primitive for thread :sub:`0` using
    addition (+) as the reduction operator.

    Returns a callable object that can be linked to and invoked from
    device code. It can be invoked with the following signatures:

    - `(item: dtype) -> dtype)`: Each thread contributes a single item to
      the reduction.
    - `(items: numba.types.Array) -> dtype`: Each thread contributes an
      array of items to the reduction. The array must contain at least
      `items_per_thread` items; only the first `items_per_thread` items
      are included in the reduction.
    - `(item: dtype, num_valid: int) -> dtype`: The first `num_valid`
      threads contribute a single item to the reduction. Items from all
      other threads are ignored.

    Args:
        dtype: Data type being reduced
        threads_per_block: Number of threads in a block. Can be an
            integer or a tuple of 2 or 3 integers.
        items_per_thread: The number of items each thread owns
        algorithm: Algorithm to use for the reduction (one of "raking",
            "raking_commutative_only", "warp_reductions")
        methods: A dict of methods for user-defined types

    Warning:
        The return value is undefined in threads other than thread :sub:`0`.

    Example:
        The code snippet below illustrates a sum of 128 integer items
        partitioned across 128 threads.

        .. literalinclude:: ../../python/cuda_cccl/tests/coop/test_block_reduce_api.py
            :language: python
            :dedent:
            :start-after: example-begin imports
            :end-before: example-end imports

        The following snippet shows how to invoke the returned
        ``block_sum`` primitive:

        .. literalinclude:: ../../python/cuda_cccl/tests/coop/test_block_reduce_api.py
            :language: python
            :dedent:
            :start-after: example-begin sum
            :end-before: example-end sum

        Suppose the set of inputs across the block of threads is
        ``{ 1, 1, 1, 1, ..., 1 }``. The output in thread :sub:`0`
        is ``{ 128 }``.
    """
    return _reduce(
        dtype=dtype,
        threads_per_block=threads_per_block,
        items_per_thread=items_per_thread,
        binary_op=None,
        algorithm=algorithm,
        methods=methods,
    )



# Copyright (c) 2025, NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
#
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

"""
cuda.coop._scan_op
==================

This module implements the ``ScanOp`` class and related functions.
"""

import operator
from enum import Enum

import numpy as np

from ._typing import (
    ScanOpType,
)


class ScanOpCategory(Enum):
    """
    Represents the category of a scan operator.  This is used to guide
    specialization of the C++ API.
    """

    Sum = "Sum"
    """
    Represents a sum operator.
    """

    Known = "Known"
    """
    Represents one of the known, non-sum associative operators: multiply,
    minimum, maximum, bitwise AND, bitwise OR, and bitwise XOR.
    """

    Callable = "Callable"
    """
    Represents a user-defined callable operator (e.g. a Python function).
    """


CUDA_STD_PLUS = "::cuda::std::plus<T>"
CUDA_STD_MULTIPLIES = "::cuda::std::multiplies<T>"
CUDA_STD_BIT_AND = "::cuda::std::bit_and<T>"
CUDA_STD_BIT_OR = "::cuda::std::bit_or<T>"
CUDA_STD_BIT_XOR = "::cuda::std::bit_xor<T>"
# `std::min` and `std::max` are STL functions, not operators, so there aren't
# `::cuda`-prefixed versions.  Instead, CUDA provides `::cuda::minimum<T>` and
# `::cuda::maximum<T>`, which *are* function objects, which is what we want.
CUDA_MINIMUM = "::cuda::minimum<T>"
CUDA_MAXIMUM = "::cuda::maximum<T>"


class ScanOp:
    """
    Represents an associative binary operator for a prefix scan operation.
    This helper class is used to *normalize* the operator provided by a user,
    which may be a string, NumPy/Python operator, or callable, to a form that
    can be used in Numba CUDA JIT'd kernels.
    """

    # Set of all ops interpreted as sum (for (inclusive|exclusive)_sum).
    SUM_OPS = {
        "+",
        "add",
        "plus",
        np.add,
        operator.add,
    }
    """
    Set of all ops interpreted as sum (for (inclusive|exclusive)_sum).
    """

    # Map of all known (non-sum) operators to their C++ type representations.
    KNOWN_OPS = {
        # String names
        "mul": CUDA_STD_MULTIPLIES,
        "multiplies": CUDA_STD_MULTIPLIES,
        "min": CUDA_MINIMUM,
        "minimum": CUDA_MINIMUM,
        "max": CUDA_MAXIMUM,
        "maximum": CUDA_MAXIMUM,
        "bit_and": CUDA_STD_BIT_AND,
        "bit_or": CUDA_STD_BIT_OR,
        "bit_xor": CUDA_STD_BIT_XOR,
        # String operators
        "*": CUDA_STD_MULTIPLIES,
        "&": CUDA_STD_BIT_AND,
        "|": CUDA_STD_BIT_OR,
        "^": CUDA_STD_BIT_XOR,
        # NumPy functions
        np.maximum: CUDA_MAXIMUM,
        np.minimum: CUDA_MINIMUM,
        np.multiply: CUDA_STD_MULTIPLIES,
        np.bitwise_and: CUDA_STD_BIT_AND,
        np.bitwise_or: CUDA_STD_BIT_OR,
        np.bitwise_xor: CUDA_STD_BIT_XOR,
        # Python operator module functions.
        operator.mul: CUDA_STD_MULTIPLIES,
        operator.and_: CUDA_STD_BIT_AND,
        operator.or_: CUDA_STD_BIT_OR,
        operator.xor: CUDA_STD_BIT_XOR,
    }
    """
    Map of all known (non-sum) operators to their C++ type representations.
    """

    def __init__(self, op: ScanOpType):
        """
        Initializes the ScanOp instance.

        :param op: Supplies the :ref:`ScanOpType` scan operator to use
            for the block-wide scan.
        :type op: ScanOpType

        :raises ValueError: If the provided operator is not supported.

        """
        if isinstance(op, ScanOp):
            # If op is already a ScanOp instance, just prime this instance
            # with its values and return.
            self.op = op.op
            self.op_category = op.op_category
            self.op_cpp = op.op_cpp
            return

        self.op = op
        self.op_category = None
        self.op_cpp = None

        # Handle string names and operators.
        if isinstance(op, str):
            if op in self.SUM_OPS:
                self.op_category = ScanOpCategory.Sum
                self.op_cpp = CUDA_STD_PLUS
            elif op in self.KNOWN_OPS:
                self.op_category = ScanOpCategory.Known
                self.op_cpp = self.KNOWN_OPS[op]
            else:
                raise ValueError(f"Unsupported scan operator: {op}")

        # Handle NumPy functions or other callables.
        elif callable(op):
            if op in self.SUM_OPS:
                self.op_category = ScanOpCategory.Sum
                self.op_cpp = CUDA_STD_PLUS
            elif op in self.KNOWN_OPS:
                self.op_category = ScanOpCategory.Known
                self.op_cpp = self.KNOWN_OPS[op]
            else:
                # Custom callable; no op_cpp representation.
                self.op_category = ScanOpCategory.Callable
        else:
            raise ValueError(f"Unsupported scan op type: {type(op)}")

    def __repr__(self):
        return f"ScanOp({self.op})"

    @property
    def is_sum(self):
        """
        Returns ``True`` if the scan operator is a sum operator.
        """
        return self.op_category == ScanOpCategory.Sum

    @property
    def is_known(self):
        """
        Returns ``True`` if the scan operator is a known operator.
        """
        return self.op_category == ScanOpCategory.Known

    @property
    def is_callable(self):
        """
        Returns ``True`` if the scan operator is a callable.
        """
        return self.op_category == ScanOpCategory.Callable



# Copyright (c) 2024, NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
#
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

from typing import TYPE_CHECKING, Tuple, Union

import numba

from .._common import (
    CUB_BLOCK_SCAN_ALGOS,
    CudaSharedMemConfig,
    dim3,
    make_binary_tempfile,
    normalize_dim_param,
    normalize_dtype_param,
)
from .._types import (
    Algorithm,
    Dependency,
    DependentArray,
    Invocable,
    Pointer,
    TemplateParameter,
    Value,
)

if TYPE_CHECKING:
    import numpy as np


TEMPLATE_PARAMETERS = [
    TemplateParameter("KeyT"),
    TemplateParameter("BLOCK_DIM_X"),
    TemplateParameter("ITEMS_PER_THREAD"),
    TemplateParameter("ValueT"),
    TemplateParameter("RADIX_BITS"),
    TemplateParameter("MEMOIZE_OUTER_SCAN"),
    TemplateParameter("INNER_SCAN_ALGORITHM"),
    TemplateParameter("SMEM_CONFIG"),
    TemplateParameter("BLOCK_DIM_Y"),
    TemplateParameter("BLOCK_DIM_Z"),
]


METHOD_PARAMETERS_VARIANTS = [
    [
        Pointer(numba.uint8),
        DependentArray(Dependency("KeyT"), Dependency("ITEMS_PER_THREAD")),
    ],
    [
        Pointer(numba.uint8),
        DependentArray(Dependency("KeyT"), Dependency("ITEMS_PER_THREAD")),
        Value(numba.int32),
        Value(numba.int32),
    ],
]


# N.B. In order to support multi-dimensional block dimensions, we have to
#      defaults for all the template parameters preceding the final Y and
#      Z dimensions.  This will be improved in the future, allowing users
#      to provide overrides for the default values.

TEMPLATE_PARAMETER_DEFAULTS = {
    "ValueT": "::cub::NullType",  # Indicates keys-only sort
    "RADIX_BITS": 4,
    "MEMOIZE_OUTER_SCAN": "true",
    "INNER_SCAN_ALGORITHM": CUB_BLOCK_SCAN_ALGOS["warp_scans"],
    "SMEM_CONFIG": str(CudaSharedMemConfig.BankSizeFourByte),
}


def _get_template_parameter_specializations(
    dtype: numba.types.Type, dim: dim3, items_per_thread: int
) -> dict:
    """
    Returns a dictionary of template parameter specializations for the block
    radix sort algorithm.

    Args:
        dtype: Supplies the Numba data type.

        dim: Supplies the block dimensions.

        items_per_thread: Supplies the number of items each thread owns.

    Returns:
        A dictionary of template parameter specializations.
    """
    specialization = {
        "KeyT": dtype,
        "BLOCK_DIM_X": dim[0],
        "ITEMS_PER_THREAD": items_per_thread,
        "BLOCK_DIM_Y": dim[1],
        "BLOCK_DIM_Z": dim[2],
    }

    specialization.update(TEMPLATE_PARAMETER_DEFAULTS)

    return specialization


def _radix_sort(
    dtype: Union[str, type, "np.dtype", "numba.types.Type"],
    threads_per_block: Union[int, Tuple[int, int], Tuple[int, int, int], dim3],
    items_per_thread: int,
    descending: bool,
) -> Invocable:
    dim = normalize_dim_param(threads_per_block)
    dtype = normalize_dtype_param(dtype)

    method_name = "SortDescending" if descending else "Sort"
    template = Algorithm(
        "BlockRadixSort",
        method_name,
        "block_radix_sort",
        ["cub/block/block_radix_sort.cuh"],
        TEMPLATE_PARAMETERS,
        METHOD_PARAMETERS_VARIANTS,
    )
    specialization = template.specialize(
        _get_template_parameter_specializations(dtype, dim, items_per_thread)
    )
    return Invocable(
        temp_files=[
            make_binary_tempfile(ltoir, ".ltoir")
            for ltoir in specialization.get_lto_ir()
        ],
        temp_storage_bytes=specialization.temp_storage_bytes,
        temp_storage_alignment=specialization.temp_storage_alignment,
        algorithm=specialization,
    )


def make_radix_sort_keys(dtype, threads_per_block, items_per_thread):
    """
    Creates an ascending block-wide radix sort primitive over a
    :ref:`blocked arrangement <flexible-data-arrangement>` of keys.

    Example:
        The code snippet below illustrates a sort of 512 integer keys
        partitioned in a
        :ref:`blocked arrangement <flexible-data-arrangement>` across
        128 threads where each thread owns 4 consecutive keys.

        .. literalinclude:: ../../python/cuda_cccl/tests/coop/test_block_radix_sort_api.py
            :language: python
            :dedent:
            :start-after: example-begin imports
            :end-before: example-end imports

        The following snippet shows how to invoke the returned
        ``block_radix_sort`` primitive:

        .. literalinclude:: ../../python/cuda_cccl/tests/coop/test_block_radix_sort_api.py
            :language: python
            :dedent:
            :start-after: example-begin radix-sort
            :end-before: example-end radix-sort

        Suppose the set of input ``thread_keys`` across the block of
        threads is
        ``{ [511, 510, 509, 508], [507, 506, 505, 504], ..., [3, 2, 1, 0] }``.
        The corresponding output ``thread_keys`` in those threads will be
        ``{ [0, 1, 2, 3], [4, 5, 6, 7], ..., [508, 509, 510, 511] }``.

    Args:
        dtype: Data type of the keys to be sorted

        threads_per_block: The number of threads in a block, either an integer
            or a tuple of 2 or 3 integers

        items_per_thread: The number of items each thread owns

    Returns:
        A callable object that can be linked to and invoked from a CUDA kernel
    """
    return _radix_sort(dtype, threads_per_block, items_per_thread, descending=False)


def make_radix_sort_keys_descending(dtype, threads_per_block, items_per_thread):
    """
    Creates a descending block-wide radix sort primitive over a
    :ref:`blocked arrangement <flexible-data-arrangement>` of keys.

    Example:
        The code snippet below illustrates a sort of 512 integer keys
        partitioned in a
        :ref:`blocked arrangement <flexible-data-arrangement>` across
        128 threads where each thread owns 4 consecutive keys.

        .. literalinclude:: ../../python/cuda_cccl/tests/coop/test_block_radix_sort_api.py
            :language: python
            :dedent:
            :start-after: example-begin imports
            :end-before: example-end imports

        The following snippet shows how to invoke the returned
        ``block_radix_sort`` primitive:

        .. literalinclude:: ../../python/cuda_cccl/tests/coop/test_block_radix_sort_api.py
            :language: python
            :dedent:
            :start-after: example-begin radix-sort-descending
            :end-before: example-end radix-sort-descending

        Suppose the set of input ``thread_keys`` across the block of
        threads is
        ``{ [0, 1, 2, 3], [4, 5, 6, 7], ..., [508, 509, 510, 511] }``.
        The corresponding output ``thread_keys`` in those threads will be
        ``{ [511, 510, 509, 508], [507, 506, 505, 504], ..., [3, 2, 1, 0] }``.

    Args:
        dtype: Data type of the keys to be sorted

        threads_per_block: The number of threads in a block, either an integer
            or a tuple of 2 or 3 integers

        items_per_thread: The number of items each thread owns

    Returns:
        A callable object that can be linked to and invoked from a CUDA kernel
    """
    return _radix_sort(dtype, threads_per_block, items_per_thread, descending=True)
